from pydmclab.core.comp import CompTools
from pydmclab.core.struc import StrucTools
from pydmclab.utils.handy import read_json, write_json
from pydmclab.data.thermochem import (
    mus_at_0K,
    mus_at_T,
    mp2020_compatibility_dmus,
    mus_from_mp_no_corrections,
)
from pydmclab.data.features import atomic_masses

import os
import numpy as np
from scipy.spatial import ConvexHull
from scipy.optimize import minimize
import multiprocessing as multip
import math
from itertools import combinations

HERE = os.path.dirname(os.path.abspath(__file__))

"""
Contents:

    1) Classes allowing you to compute decomposition energies using the convex hull analysis
        Standard approach:
            - collect formation energies for all compounds in given chemical system(s)
                - using your own calculations, MPQuery, or a combination
            - map these formation energies into various phase diagrams
                - w/ GetHullInputData
            - compute convex hulls for each phase diagram
                - w/ AnalyzeHull
            - if you are doing this for many compounds / chemical spaces, you want to parallelize it
                - w/ ParallelHullAnalysis

"""


class GetHullInputData(object):
    """
    Generates hull-relevant data
    Designed to be executed once all compounds and ground-state formation energies are known
    """

    def __init__(self, compound_to_energy, formation_energy_key):
        """
        Args:
            compound_to_energy (dict) - {formula (str) :
                                            {formation_energy_key (str) :
                                                formation energy (float, eV/atom)}}
                - it's ok to have other keys in this dictionary, just have to have a formation_energy_key that maps to formation energies
                - these formation energies can be 0 K, at some temperature, or computed in any way that you want, they just have to be compatible w/ one another
            formation_energy_key (str) - key within compound_to_energy to use for formation energy

        Returns:
            dictionary of {formula (str) : formation energy (float, eV/atom)}
                - note that elements are removed from this dictionary, they will be added back in later
                    - this is a legacy thing that is kind of tedious but (probably) needs to be done
        """
        self.compound_to_energy = {
            CompTools(k).clean: compound_to_energy[k][formation_energy_key]
            for k in compound_to_energy
            if len(CompTools(k).els) > 1
        }

    @property
    def compounds(self):
        """
        Returns:
            list of compounds (str)
        """
        return list(self.compound_to_energy.keys())

    @property
    def chemical_spaces_and_subspaces(self):
        """
        Returns:
            list of unique chemical spaces (tuple)
                - a chemical space would be (Ca, O, Ti) for CaTiO3
        """
        compounds = self.compounds
        return list(set([tuple(CompTools(c).els) for c in compounds]))

    @property
    def chemical_subspaces(self):
        """
        Returns:
            list of unique chemical spaces (tuple) that do not define convex hull spaces
                (Ca, O, Ti) is the space of CaTiO3 and Ca2TiO4
                    if CaTiO3 and CaO are found in compound_to_energy, (Ca, O) is a subspace because it does not include CaTiO3
        """
        all_spaces = self.chemical_spaces_and_subspaces
        subspaces = [
            all_spaces[i]
            for i in range(len(all_spaces))
            for j in range(len(all_spaces))
            if set(all_spaces[i]).issubset(all_spaces[j])
            if all_spaces[i] != all_spaces[j]
        ]
        return list(set(subspaces))

    def hull_spaces(self, fjson=False, remake=False, write=False):
        """
        this can take a while to compute, so sometimes it's useful to write it to a json file

        Args:
            fjson (str) - file name to write hull spaces to
            remake (bool) - if True, repeat analysis; if False, read json
            write (bool) - if True, write json; if False, don't write json

        Returns:
            list of unique chemical spaces (set) that do define convex hull spaces
                these are the convex hulls that must be computed
                so if you are working with binaries and ternaries, but no quaternaries, hull spaces will be maximally 3 dimensions (elements)
        """
        if not fjson:
            fjson = "hull_spaces.json"
        if not remake and os.path.exists(fjson):
            d = read_json(fjson)
            return d["hull_spaces"]
        chemical_spaces_and_subspaces = self.chemical_spaces_and_subspaces
        chemical_subspaces = self.chemical_subspaces
        d = {
            "hull_spaces": [
                s
                for s in chemical_spaces_and_subspaces
                if s not in chemical_subspaces
                if len(s) > 1
            ]
        }
        if write:
            d = write_json(d, fjson)
        return d["hull_spaces"]

    def hullin_data(self, fjson=False, remake=True):
        """

        Args:
            fjson (str) - file name to write hull data to
            remake (bool) - if True, run analysis + write json; if False, read json

        Returns:
            dict of {chemical space (str) : {formula (str) : {'E' : formation energy (float),
                                                              'amts' : {el (str) : fractional amt of el in formula (float) for el in space}}
                                            for all relevant formulas including elements}
                - elements are automatically given formation energy = 0
                - chemical space is now in 'el1_el2_...' format to be jsonable
                - each "chemical space" is a convex hull that must be computed
        """
        if not fjson:
            fjson = "hull_input_data.json"
        if os.path.exists(fjson) and not remake:
            return read_json(fjson)

        hullin_data = {}
        hull_spaces = self.hull_spaces()
        compounds = self.compounds
        compound_to_energy = self.compound_to_energy
        for space in hull_spaces:
            for el in space:
                compound_to_energy[el] = 0
            relevant_compounds = [
                c for c in compounds if set(CompTools(c).els).issubset(set(space))
            ] + list(space)
            hullin_data["_".join(list(space))] = {
                c: {
                    "E": compound_to_energy[c],
                    "amts": {el: CompTools(c).mol_frac(el=el) for el in space},
                }
                for c in relevant_compounds
            }
        write_json(hullin_data, fjson)
        return read_json(fjson)


class AnalyzeHull(object):
    """
    Determines stability for one chemical space (hull)
    Designed to be parallelized over chemical spaces
    Output can be a dictionary with hull results for one chemical space or can be run for specified compounds
    """

    def __init__(self, hullin_data, chemical_space, exclude_els=False):
        """
        Args:
            hullin_data (dict) - dictionary generated in GetHullInputData().hullin_data
            chemical_space (str) - chemical space to analyze in 'el1_el2_...' (alphabetized) format
                - to compute Ca-Ti-O phase diagram, chemical_space would be 'Ca_O_Ti'
        Returns:

            hullin_data (dict) - dictionary of {compound (str) :
                                                    {'E' : formation energy (float),
                                                     'amts' :
                                                        {el (str) :
                                                            fractional amt of el in compound (float)
                                                                for el in chemical_space}}}})}

                - grabs only the relevant sub-dict from hullin_data

            chemical_space (tuple) - chemical space to analyze
                - changes chemical space to tuple (el1, el2, ...)

            exclude_els (bool) - if True, the energies of the elements will be made very high
                - this is useful when computing mixing energies (e.g., for voltage curves)

        """
        hullin_data = hullin_data[chemical_space]

        keys_to_remove = [k for k in hullin_data if CompTools(k).n_els == 1]
        hullin_data = {
            k: hullin_data[k] for k in hullin_data if k not in keys_to_remove
        }

        els = chemical_space.split("_")

        for el in els:
            hullin_data[el] = {
                "E": 0 if not exclude_els else 0.1,
                "amts": {
                    els[i]: CompTools(el).mol_frac(els[i]) for i in range(len(els))
                },
            }

        self.hullin_data = hullin_data

        self.chemical_space = tuple(els)

    @property
    def sorted_compounds(self):
        """
        Returns:
            alphabetized list of compounds (str) in specified chemical space
        """
        return sorted(list(self.hullin_data.keys()))

    def amts_matrix(self, compounds="all", chemical_space="all"):
        """
        Args:
            compounds (str or list) - if 'all', use all compounds; else use specified list
                - note: this gets modified for you as needed to minimize cpu time
            chemical_space - if 'all', use entire space; else use specified tuple
                - note: this gets modified for you as needed to minimize cpu time

        Returns:
            matrix (2D array) with the fractional composition of each element in each compound (float)
                - each row is a different compound (ordered going down alphabetically)
                - each column is a different element (ordered across alphabetically)
        """
        if chemical_space == "all":
            chemical_space = self.chemical_space
        hullin_data = self.hullin_data
        if compounds == "all":
            compounds = self.sorted_compounds
        A = np.zeros((len(compounds), len(chemical_space)))
        for row in range(len(compounds)):
            compound = compounds[row]
            for col in range(len(chemical_space)):
                el = chemical_space[col]
                A[row, col] = hullin_data[compound]["amts"][el]
        return A

    def formation_energy_array(self, compounds="all"):
        """
        Args:
            compounds (str or list) - if 'all', use all compounds; else use specified list
                - this gets modified for you as needed to minimize cpu time

        Returns:
            1D array of formation energies (float) for each compound ordered alphabetically
        """
        hullin_data = self.hullin_data
        if compounds == "all":
            compounds = self.sorted_compounds
        return np.array([hullin_data[c]["E"] for c in compounds])

    def hull_input_matrix(self, compounds="all", chemical_space="all"):
        """
        Args:
            compounds (str or list) - if 'all', use all compounds; else use specified list
                - this gets modified for you as needed to minimize cpu time
            chemical_space - if 'all', use entire space; else use specified tuple
                - this gets modified for you as needed to minimize cpu time
        Returns:
            amts_matrix, but replacing the last column with the formation energy
                - this is because convex hulls are defined by (n-1) composition axes
                    - e.g., in a A-B phase diagram, specifying the fractional composition of A sets the composition of B
        """
        A = self.amts_matrix(compounds, chemical_space)
        b = self.formation_energy_array(compounds)
        X = np.zeros(np.shape(A))
        for row in range(np.shape(X)[0]):
            for col in range(np.shape(X)[1] - 1):
                X[row, col] = A[row, col]
            X[row, np.shape(X)[1] - 1] = b[row]
        return X

    @property
    def hull(self):
        """
        Returns:
            scipy.spatial.ConvexHull object for all compounds in a chemical space
        """
        return ConvexHull(self.hull_input_matrix(compounds="all", chemical_space="all"))

    @property
    def hull_points(self):
        """
        Returns:
            array of (composition, formation energy) points (2-element tuple) fed to ConvexHull
        """
        return self.hull.points

    @property
    def hull_vertices(self):
        """
        Returns:
            array of indices (int) in hull_points corresponding with the points that are on the hull
        """
        return self.hull.vertices

    @property
    def hull_simplices(self):
        """
        Returns:
            indices of points forming the simplical facets of the convex hull.
        """
        return self.hull.simplices

    @property
    def stable_compounds(self):
        """
        Returns:
            list of compounds (str) that correspond with vertices on the hull
                - these are stable compounds
        """
        hullin_data = self.hullin_data
        hull_vertices = self.hull_vertices
        compounds = self.sorted_compounds
        return [
            compounds[i] for i in hull_vertices if hullin_data[compounds[i]]["E"] <= 0
        ]

    @property
    def unstable_compounds(self):
        """
        Args:

        Returns:
            list of compounds that do not correspond with vertices (str)
                - these are "above" the hull
        """
        compounds = self.sorted_compounds
        stable_compounds = self.stable_compounds
        return [c for c in compounds if c not in stable_compounds]

    def competing_compounds(self, compound):
        """
        Args:
            compound (str) - the compound (str) to analyze

        Returns:
            list of compounds (str) that may participate in the decomposition reaction for the input compound
                - these must have the same elements as the input compound or a subset of those elements
        """
        compounds = self.sorted_compounds
        if compound in self.unstable_compounds:
            compounds = self.stable_compounds
        competing_compounds = [
            c
            for c in compounds
            if c != compound
            if set(CompTools(c).els).issubset(CompTools(compound).els)
        ]
        return competing_compounds

    def A_for_decomp_solver(self, compound, competing_compounds):
        """
        Args:
            compound (str) - the compound (str) to analyze
            competing_compounds (list) - list of compounds (str) that may participate in the decomposition reaction for the input compound

        Returns:
            matrix (2D array) of elemental amounts (float) used for implementing molar conservation during decomposition solution
                - i.e., a decomposition reaction has to conserve the total number of each element on both sides of the reaction
        """
        chemical_space = tuple(CompTools(compound).els)
        atoms_per_fu = [CompTools(c).n_atoms for c in competing_compounds]
        A = self.amts_matrix(competing_compounds, chemical_space)
        for row in range(len(competing_compounds)):
            for col in range(len(chemical_space)):
                A[row, col] *= atoms_per_fu[row]
        return A.T

    def b_for_decomp_solver(self, compound):
        """
        Args:
            compound (str) - the compound (str) to analyze

        Returns:
            array of elemental amounts (float) used for implementing molar conservation during decomposition solution
        """
        chemical_space = tuple(CompTools(compound).els)
        return np.array([CompTools(compound).stoich(el) for el in chemical_space])

    def Es_for_decomp_solver(self, competing_compounds):
        """
        Args:
            competing_compounds (list) - list of compounds (str) that may participate in the decomposition reaction for the input compound

        Returns:
            array of formation energies per formula unit (float) used for minimization problem during decomposition solution
                - these have to be Ef per f.u. because we are computing reaction energies
        """
        atoms_per_fu = [CompTools(c).n_atoms for c in competing_compounds]
        Es_per_atom = self.formation_energy_array(competing_compounds)
        return [
            Es_per_atom[i] * atoms_per_fu[i] for i in range(len(competing_compounds))
        ]

    def decomp_solution(self, compound):
        """
        Args:
            compound (str) - the compound (str) to analyze

        Returns:
            scipy.optimize.minimize result
                for finding the linear combination of competing compounds that minimizes the competing formation energy
                    - i.e., the fractional composition of decomposition products leading to the most positive deecomposition energy
        """
        competing_compounds = self.competing_compounds(compound)
        A = self.A_for_decomp_solver(compound, competing_compounds)
        b = self.b_for_decomp_solver(compound)
        Es = self.Es_for_decomp_solver(competing_compounds)
        n0 = [0 for c in competing_compounds]
        max_bound = CompTools(compound).n_atoms
        bounds = [(0, max_bound) for c in competing_compounds]

        def competing_formation_energy(nj):
            nj = np.array(nj)
            return np.dot(nj, Es)

        constraints = [{"type": "eq", "fun": lambda x: np.dot(A, x) - b}]
        maxiter, disp = 1000, False
        # slowly relax tolerance in case minimize can't converge
        for tol in [1e-6, 1e-4, 1e-3, 1e-2]:
            solution = minimize(
                competing_formation_energy,
                n0,
                method="SLSQP",
                bounds=bounds,
                constraints=constraints,
                tol=tol,
                options={"maxiter": maxiter, "disp": disp},
            )
            if solution.success:
                return solution
        return solution

    def decomp_products(self, compound):
        """
        Args:
            compound (str) - the compound (str) to analyze

        Returns:
            dictionary of {competing compound (str) : {'amt' : stoich weight in decomp rxn (float),
                                                       'E' : formation energy (float)}
                                                            for all compounds in the stability-defining reaction}
                np.nan if decomposition analysis fails
        """
        hullin_data = self.hullin_data
        competing_compounds = self.competing_compounds(compound)

        if (len(competing_compounds) == 0) or (
            np.max([CompTools(c).n_els for c in competing_compounds]) == 1
        ):
            return {
                el: {"amt": CompTools(compound).stoich(el), "E": 0}
                for el in CompTools(compound).els
            }
        solution = self.decomp_solution(compound)
        if solution.success:
            resulting_amts = solution.x
        elif hullin_data[compound]["E"] > 0:
            return {
                el: {"amt": CompTools(compound).stoich(el), "E": 0}
                for el in CompTools(compound).els
            }
        else:
            # in case scipy.minimize fails to converge
            print(compound)
            print("\n\n\nFAILURE!!!!\n\n\n")
            print(compound)
            return np.nan

        # put irrelevant compounds to zero to make reaction more clear
        min_amt_to_show = 1e-4
        decomp_products = dict(zip(competing_compounds, resulting_amts))
        relevant_decomp_products = [
            k for k in decomp_products if decomp_products[k] > min_amt_to_show
        ]
        decomp_products = {
            k: {"amt": decomp_products[k], "E": hullin_data[k]["E"]}
            if k in hullin_data
            else 0
            for k in relevant_decomp_products
        }

        # here's some stuff that's probably unnecessary if everything is working properly
        """
        el_totals_to_match = {el : CompTools(compound).stoich(el) for el in CompTools(compound).els}
        free_decomp_products = [cmpd for cmpd in decomp_products if CompTools(cmpd).n_els == 1]
        
        decomp_totals = {el : 0 for el in CompTools(compound).els}
        for decomp_product in decomp_totals:
            for el in CompTools(decomp_product).els:
                decomp_totals[el] += CompTools(decomp_product).stoich(el)
        for el in free_decomp_products:
            decomp_products[el]['amt'] += el_totals_to_match[el] - decomp_totals[el]
        """
        return decomp_products

    def decomp_energy(self, compound):
        """
        Args:
            compound (str) - the compound (str) to analyze

        Returns:
            decomposition energy (float)
                - if < 0, this is the magnitude of formation energy that compound could increase and still be on the hull
                - if > 0, this is the energy above the hull
        """
        hullin_data = self.hullin_data
        decomp_products = self.decomp_products(compound)
        if isinstance(decomp_products, float):
            return np.nan
        decomp_energy = 0
        for k in decomp_products:
            decomp_energy += (
                decomp_products[k]["amt"]
                * decomp_products[k]["E"]
                * CompTools(k).n_atoms
            )
        return (
            hullin_data[compound]["E"] * CompTools(compound).n_atoms - decomp_energy
        ) / CompTools(compound).n_atoms

    @property
    def hull_output_data(self):
        """
        Returns:
            stability data (dict) for all compounds in the specified chemical space
                {compound (str) : {'Ef' : formation energy (float),
                                   'Ed' : decomposition energy (float),
                                   'rxn' : decomposition reaction (str),
                                   'stability' : stable (True) or unstable (False)}}
        """
        data = {}
        hullin_data = self.hullin_data
        compounds, stable_compounds = self.sorted_compounds, self.stable_compounds
        for c in compounds:
            if c in stable_compounds:
                stability = True
            else:
                stability = False
            Ef = hullin_data[c]["E"]
            Ed = self.decomp_energy(c)
            decomp_products = self.decomp_products(c)

            if isinstance(decomp_products, float):
                data[c] = np.nan
                continue
            decomp_rxn = [
                "_".join([str(np.round(decomp_products[k]["amt"], 4)), k])
                for k in decomp_products
            ]
            decomp_rxn = " + ".join(decomp_rxn)
            data[c] = {"Ef": Ef, "Ed": Ed, "rxn": decomp_rxn, "stability": stability}
        return data

    def cmpd_hull_output_data(self, compound):
        """
        Args:
            compound (str) - formula to get data for

        Returns:
            hull_output_data but only for single compound
                {'Ef' : formation energy (float),
                'Ed' : decomposition energy (float),
                'rxn' : decomposition reaction (str),
                'stability' : stable (True) or unstable (False)}

            - so you could loop through compounds to get this or do things on a chemical space basis
        """
        data = {}
        hullin_data = self.hullin_data
        stable_compounds = self.stable_compounds
        c = compound
        if c in stable_compounds:
            stability = True
        else:
            stability = False
        Ef = hullin_data[c]["E"]
        Ed = self.decomp_energy(c)
        decomp_products = self.decomp_products(c)
        if isinstance(decomp_products, float):
            return {c: np.nan}
        decomp_rxn = [
            "_".join([str(np.round(decomp_products[k]["amt"], 4)), k])
            for k in decomp_products
        ]
        decomp_rxn = " + ".join(decomp_rxn)
        data[c] = {"Ef": Ef, "Ed": Ed, "rxn": decomp_rxn, "stability": stability}
        return data[c]


class ParallelHulls(object):
    """
    this allows you to parallelize the hullin_data and hullout_data generation over cores using multiprocessing

    """

    def __init__(
        self,
        compound_to_energy,
        formation_energy_key="Ef_mp",
        n_procs=4,
        data_dir=False,
        fresh_restart=False,
    ):
        """
         compound_to_energy (dict) - {formula (str) :
                                        {formation_energy_key (str) :
                                            formation energy (float, eV/atom)}}
            - it's ok to have other keys in this dictionary, just have to have a formation_energy_key that maps to formation energies
            - these formation energies can be 0 K, at some temperature, or computed in any way that you want, they just have to be compatible w/ one another
        formation_energy_key (str) - key within compound_to_energy to use for formation energy
        n_procs (int) - number of cores to use
            - if 'all', will use every core availble minus 1
        data_dir (str) - directory to save data to (if False, writes to os.getcwd())
        fresh_restart (bool) - if True, will re-compute everything from scratch
        """
        self.n_procs = n_procs if n_procs != "all" else multip.cpu_count() - 1
        if not data_dir:
            self.data_dir = os.getcwd()
        else:
            self.data_dir = data_dir
        self.fresh_restart = fresh_restart

        self.compound_to_energy = {
            k: compound_to_energy[k][formation_energy_key]
            for k in compound_to_energy
            if len(CompTools(k).els) > 1
        }

    @property
    def compounds(self):
        """
        Returns:
            list of compounds (str)
        """
        compounds = list(self.compound_to_energy.keys())
        return sorted(list([c for c in compounds if CompTools(c).n_els > 1]))

    @property
    def chemical_spaces(self):
        """
        Returns:
            list of unique chemical spaces (tuple)
                - a chemical space would be (Ca, O, Ti) for CaTiO3
        """
        compounds = self.compounds
        chemical_spaces = []
        for c in compounds:
            space = "_".join(sorted(list(set(CompTools(c).els))))
            chemical_spaces.append(space)
        return sorted(list(set(chemical_spaces)))

    def hullin_from_space(self, space, verbose=True):
        """
        Function called at each parallelized generation of hull input file (same as GetHullInputData.hullin_data)

        Args:
            space (str) - '_'.join(elements) (str) in chemical space
            verbose (bool) - print space or not
        Returns:
            {compound (str) : {'E' : formation energy (float, eV/atom),
                        'amts' : {el (str) : fractional amount of el in formula (float)}}}
        """
        compound_to_energy = self.compound_to_energy
        if verbose:
            print(space)
        space = space.split("_")
        for el in space:
            compound_to_energy[el] = 0
        relevant_compounds = [
            c for c in compound_to_energy if set(CompTools(c).els).issubset(set(space))
        ] + list(space)
        return {
            c: {
                "E": compound_to_energy[c],
                "amts": {el: CompTools(c).mol_frac(el=el) for el in space},
            }
            for c in relevant_compounds
        }

    def parallel_hullin(self, fjson=False, verbose=True):
        """
        Parallel generation of hull input data (same as GetHullInputData.hullin_data)

        Args:
            fjson (os.PathLike) - path to write dictionary of hull input data
            verbose (bool) - print space or not
        Returns:
            {chemical space (str) :
                {compound (str) :
                    {'E' : formation energy (float, eV/atom),
                    'amts' : {el (str) : fractional amount of el in formula (float)}}}}
        """
        remake = self.fresh_restart
        chemical_spaces = self.chemical_spaces
        if not fjson:
            fjson = os.path.join(self.data_dir, "hull_input_data.json")
        if (remake == True) or not os.path.exists(fjson):
            hullin_data = {}
            pool = multip.Pool(processes=self.n_procs)
            results = [
                r
                for r in pool.starmap(
                    self.hullin_from_space,
                    [(space, verbose) for space in chemical_spaces],
                )
            ]
            pool.close()
            hullin_data = dict(zip(chemical_spaces, results))
            return write_json(hullin_data, fjson)
        else:
            return read_json(fjson)

    def smallest_space(self, hullin, formula, verbose=False):
        """
        function to parallelize finding the smallest hull to compute for each compound

        Args:
            hullin (dict) - {space (str, '_'.join(elements)) :
                                {formula (str) :
                                    {'E' : formation energy (float, eV/atom),
                                    'amts' :
                                        {element (str) : fractional amount of element in formula (float)}
                                    }
                                }
                            }
            formula (str) - chemical formula
            verbose (bool) - print formula or not

        Returns:
            chemical space (str, '_'.join(elements)) of the convex hull that is easiest to compute for a given compound
                - for instance, if I want the decomp energy of CaO, there's no need to do the Ca-Ti-O hull, only the Ca-O one
        """
        if verbose:
            print(formula)
        spaces = sorted(list(hullin.keys()))
        relevant = [s for s in spaces if formula in hullin[s]]
        sizes = [s.count("_") for s in relevant]
        small = [relevant[i] for i in range(len(sizes)) if sizes[i] == np.min(sizes)]
        sizes = [len(hullin[s]) for s in small]
        smallest = [small[i] for i in range(len(small)) if sizes[i] == np.min(sizes)]
        return smallest[0]

    def smallest_spaces(self, hullin, fjson=False, verbose=False):
        """
        parallel generation of smallest spaces for each compound

        Args:
            hullin (dict) - same as smallest_space
            fjson (os.PathLike) - path to write dictionary of smallest spaces
            verbose (bool) - print formula or not

        Returns:
            {formula (str) :
                chemical space (str, '_'.join(elements)) of convex hull that is easiest to compute}
        """
        remake = self.fresh_restart
        compounds = self.compounds
        if not fjson:
            fjson = os.path.join(self.data_dir, "smallest_spaces.json")
        if not remake and os.path.exists(fjson):
            return read_json(fjson)
        pool = multip.Pool(processes=self.n_procs)
        smallest = [
            r
            for r in pool.starmap(
                self.smallest_space,
                [(hullin, compound, verbose) for compound in compounds],
            )
        ]
        pool.close()
        data = dict(zip(compounds, smallest))
        return write_json(data, fjson)

    def compound_stability(self, hullin, smallest_spaces, formula, verbose=False):
        """
        Args:
            hullin (dict) - hull input dictionary
            smallest_spaces (dict) - {formula (str) : smallest chemical space having formula (str)}
            formula (str) - chemical formula
            verbose (bool) - print formula or not

        Returns:
            {'Ef' : formation energy (float, eV/atom),
             'Ed' : decomposition energy (float, eV/atom),
             'rxn' : decomposition reaction (str),
             'stability' : bool (True if on hull)}
        """
        if verbose:
            print(formula)
        if CompTools(formula).n_els == 1:
            return {"Ef": 0, "Ed": 0, "stability": True, "rxn": "1_%s" % formula}
        space = smallest_spaces[formula]
        return AnalyzeHull(hullin, space).cmpd_hull_output_data(formula)

    def parallel_hullout(
        self,
        hullin,
        smallest_spaces,
        compounds="all",
        fjson=False,
        remake=False,
        verbose=False,
    ):
        """
        Args:
            hullin (dict) - hull input dictionary
            smallest_spaces (dict) - {formula (str) : smallest chemical space having formula (str)}
            compounds (str or list) - if 'all', use every compound, else specify which compounds to compute
            fjson (os.PathLike) - path to write dictionary of hull output data
            reemake (bool) - remake the data or not
            verbose (bool) - print formula or not

        Returns:
            {'Ef' : formation energy (float, eV/atom),
             'Ed' : decomposition energy (float, eV/atom),
             'rxn' : decomposition reaction (str),
             'stability' : bool (True if on hull)}
        """
        if not fjson:
            fjson = os.path.join(self.data_dir, "hullout.json")
        if not remake and os.path.exists(fjson):
            return read_json(fjson)
        pool = multip.Pool(processes=self.n_procs)
        if compounds == "all":
            compounds = sorted(list(smallest_spaces.keys()))
        results = [
            r
            for r in pool.starmap(
                self.compound_stability,
                [
                    (hullin, smallest_spaces, compound, verbose)
                    for compound in compounds
                ],
            )
        ]
        pool.close()
        data = dict(zip(compounds, results))
        return write_json(data, fjson)



from pydmclab.hpc.vasp import VASPSetUp
from pydmclab.hpc.analyze import AnalyzeVASP
from pydmclab.utils.handy import read_yaml, write_yaml
from pydmclab.data.configs import (
    load_vasp_configs,
    load_slurm_configs,
    load_sub_configs,
    load_partition_configs,
)

from pymatgen.core.structure import Structure

import os
from shutil import copyfile, rmtree
import subprocess
import warnings

HERE = os.path.dirname(os.path.abspath(__file__))


class SubmitTools(object):
    """
    This class is focused on figuring out how to prepare chains of calculations
        - the idea being that the output from this class is some file that you can
            "submit" to a queueing system
        - this class will automatically crawl through the VASP output files and figure out
            how to edit that submission file to finish the desired calculations

    """

    def __init__(
        self,
        launch_dir,
        final_xcs,
        magmom,
        user_configs={},
        refresh_configs=["vasp", "sub", "slurm"],
        vasp_configs_yaml=os.path.join(os.getcwd(), "_vasp_configs.yaml"),
        slurm_configs_yaml=os.path.join(os.getcwd(), "_slurm_configs.yaml"),
        sub_configs_yaml=os.path.join(os.getcwd(), "_sub_configs.yaml"),
    ):

        """
        Args:
            launch_dir (str) - directory to launch calculations from (to submit the submission file)
                - assumes initial structure is POSCAR in launch_dir
                    - LaunchTools will put it there
                - within this directory, various VASP calculation directories (calc_dirs) will be created
                        - gga-loose, gga-relax, gga-static, etc
                            - VASP will be run in each of these, but we need to run some sequentially, so we pack them together in one submission script

            final_xcs (list) - list of exchange correlation methods we want the final energy for
                - you should pass this to SubmitTools
                    - e.g., ['metagga'] if you only want metagga or ['ggau', 'metagga'] if you want both of these methods

            magmom (list) - list of magnetic moments for each atom in structure (or None if not AFM)
                - you should pass this here
                    - None if not AFM
                    - otherwise pull it from a dictionary of magmoms you made with MagTools

            user_configs (dict) - any non-default parameters you want to pass
                - these will override the defaults in the yaml files
                - look at pydmc/data/data/*configs*.yaml for the defaults
                    - note: _launch_configs.yaml will be modified using LaunchTools
                - you should be passing stuff here!
                - VASPSetUp will expect xc_to_run, calc_to_run, standard, and mag
                    - xc_to_run and calc_to_run will be passed based on final_xcs and sub_configs['packing']
                    - standard and mag will get passed to it based on launch_dir
                - you can also pass any settings in _vasp_configs.yaml, _slurm_configs.yaml, or _sub_configs.yaml here
                - _vasp_configs options:
                    see pydmc/hpc/vasp.py or pydmc/data/data/_vasp_configs.yaml for options
                - _sub_configs options:
                    vasp: vasp_std # use vasp_gam for "loose" calcs (havent implemented yet)
                    vasp_dir: /home/cbartel/shared/bin/vasp/ # where vasp executable lives
                    mpi_command: srun # how to launch on multicore/multinode (may be mpirun depending on compilation)
                    manager: '#SBATCH' # how to manage interactions with the queue (some machines dont use slurm)

                    fqueue: q.o # this will be created in the folder where you execute python

                    fresh_restart: False # True if you want to re-run all calculations; False if you want to pick up where you left off
                    force_postprocess: False # if True, run LOBSTER and Bader again if vasp_configs.lobster_static = True even if output files exist

                    files_to_inherit: ['WAVECAR', 'CONTCAR'] # usually no need to change

                    execute_flags: ['srun', 'python', 'lobster', 'bader'] # what to look for in a submission script to see if it should be launched

                    # this defines what jobs should be chained together and in what order they should be executed
                    # if you only wanted to run a static calculation for metagga, you might pass packing['metagga'] = ['metagga-static']
                    packing:
                    gga:
                    - gga-loose
                    - gga-relax
                    - gga-static
                    ggau:
                    - ggau-loose
                    - ggau-relax
                    - ggau-static
                    metagga:
                    - gga-loose
                    - gga-relax
                    - gga-static
                    - metagga-relax
                    - metagga-static
                - _slurm_configs options:
                    nodes: 1 # how many nodes per launched sequence of calcs
                    ntasks: 16 # how many total cores
                    time: 1440 # how long in minutes before hitting walltime
                    error: log.e # where to write slurm errors to in launch_dir
                    output: log.o # where to write slurm output to in launch_dir
                    account: cbartel # account to charge
                    partition: msismall # partition to use
                    job-name: # unique job name; if none provided, will default to formula.id.standard.mag.xc
                    mem: # if you need more memory specify here (e.g., 128GB)
                    constraint: # may not need this ever on MSI
                    qos: # may not need this ever on MSI

            vasp_configs_yaml (os.pathLike) - path to yaml file containing baseline vasp configs
                - there's usually no reason to change this
                - this holds some default configs for VASP
                - can always be changed with user_configs

             slurm_configs_yaml (os.pathLike) - path to yaml file containing baseline slurm configs
                - there's usually no reason to change this
                - this holds some default configs for slurm
                - can always be changed with user_configs

            sub_configs_yaml (os.pathLike) - path to yaml file containing baseline submission file configs
                - there's usually no reason to change this
                - this holds some default configs for submission files
                - can always be changed with user_configs

            refresh_configs (bool) - if True, will copy pydmc baseline configs to your local directory
                - this is useful if you've made changes to the configs files in the directory you're working in and want to start over

        Returns:
            self.launch_dir (os.pathLike) - directory to launch calculations from
            self.valid_calcs (list) - list of calculations to run
            self.slurm_configs (dotdict) - dictionary of slurm configs (in format similar to yaml)
            self.vasp_configs (dotdict) - dictionary of vasp configs (in format similar to yaml)
            self.sub_configs (dotdict) - dictionary of submission configs (in format similar to yaml)
            self.files_to_inherit (list) - list of files to copy from calc to calc
            self.structure (Structure) - pymatgen structure object from launch_dir/POSCAR
            self.magmom (list) - list of magnetic moments for each atom in structure
            self.partitions (dotdict) - dictionary of info regarding partition configurations on MSI
        """

        self.launch_dir = launch_dir

        # just a reminder of how a launch director looks
        # NOTE: should be made with LaunchTools
        top_level, unique_ID, standard, mag = launch_dir.split("/")[-4:]

        # don't run gga twice
        if ("gga" in final_xcs) and ("metagga" in final_xcs):
            warnings.warn(
                "You have both gga and metagga in final_xcs. This is probably not what you want. I'm going to remove gga from final_xcs since you get it for free w/ metagga"
            )
            final_xcs.remove("gga")

        # copy baseline vasp configs to launch_dir if they don't exist
        if not os.path.exists(vasp_configs_yaml) or ("vasp" in refresh_configs):
            _vasp_configs = load_vasp_configs()
            write_yaml(_vasp_configs, vasp_configs_yaml)

        # copy baseline slurm configs to launch_dir if they don't exist
        if not os.path.exists(slurm_configs_yaml) or ("slurm" in refresh_configs):
            _slurm_configs = load_slurm_configs()
            write_yaml(_slurm_configs, slurm_configs_yaml)

        # copy baseline sub configs to launch_dir if they don't exist
        if not os.path.exists(sub_configs_yaml) or ("sub" in refresh_configs):
            _sub_configs = load_sub_configs()
            write_yaml(_sub_configs, sub_configs_yaml)

        # we're going to modify vasp, slurm, and sub configs using one user_configs dict, so let's keep track of what's been applied
        user_configs_used = []

        # update slurm_configs based on user_configs
        slurm_configs = read_yaml(slurm_configs_yaml)
        for option in slurm_configs:
            if option in user_configs:
                if option not in user_configs_used:
                    new_value = user_configs[option]
                    slurm_configs[option] = new_value
                    user_configs_used.append(option)

        # create copy of slurm_configs to prevent unwanted updates
        self.slurm_configs = slurm_configs.copy()

        # update sub_configs based on user_configs
        sub_configs = read_yaml(sub_configs_yaml)
        for option in sub_configs:
            if option in user_configs:
                if option not in user_configs_used:
                    new_value = user_configs[option]
                    sub_configs[option] = new_value
                    user_configs_used.append(option)

        # create copy of sub_configs to prevent unwanted updates
        self.sub_configs = sub_configs.copy()

        # update vasp_configs based on user_configs
        vasp_configs = read_yaml(vasp_configs_yaml)
        for option in vasp_configs:
            if option in user_configs:
                if option not in user_configs_used:
                    new_value = user_configs[option]
                    vasp_configs[option] = new_value
                    user_configs_used.append(option)

        # determine standard and mag from launch_dir
        vasp_configs["standard"] = standard
        vasp_configs["mag"] = mag

        # include magmom in vasp_configs
        vasp_configs["magmom"] = magmom

        # create copy of vasp_configs to prevent unwanted updates
        self.vasp_configs = vasp_configs.copy()

        # need a POSCAR to initialize setup
        # LaunchTools should take care of this
        fpos = os.path.join(launch_dir, "POSCAR")
        if not os.path.exists(fpos):
            raise FileNotFoundError(
                "Need a POSCAR to initialize setup; POSCAR not found in {}".format(
                    self.launch_dir
                )
            )
        else:
            self.structure = Structure.from_file(fpos)

        # load partition configurations to help with slurm setup
        partitions = load_partition_configs()
        self.partitions = partitions

        # these are the xcs we want energies for --> each one of these should have a submission script
        # i.e., they are the end of individual chains
        self.final_xcs = final_xcs

    @property
    def queue_manager(self):
        """
        Returns queue manager (eg #SBATCH)
        """
        return self.sub_configs["manager"]

    @property
    def slurm_options(self):
        """
        Returns dictionary of slurm options
            - nodes, ntasks, walltime, etc

        To be written at the top of submission files
        """
        slurm_configs = self.slurm_configs.copy()
        options = {
            option: slurm_configs[option]
            for option in slurm_configs
            if slurm_configs[option]
        }
        partitions = self.partitions.copy()
        partition_specs = partitions[options["partition"]]

        # make sure slurm_options align with partition configurations
        if partition_specs["proc"] == "gpu":
            options["nodes"] = 1
            options["ntasks"] = 1
            options["gres"] = "gpu:%s:%s" % (
                options["partition"].split("-")[0],
                str(options["nodes"]),
            )
        if not partition_specs["sharing"]:
            options["ntasks"] = partition_specs["cores_per_node"]
        return options

    @property
    def vasp_command(self):
        """
        Returns command used to execute vasp
            e.g., 'srun -n 24 PATH_TO_VASP/vasp_std > vasp.o'
        """
        sub_configs = self.sub_configs.copy()
        vasp_configs = self.vasp_configs.copy()
        vasp_exec = os.path.join(sub_configs["vasp_dir"], sub_configs["vasp"])
        if sub_configs['mpi_command'] == 'srun':
            return "\n%s --ntasks=%s --mpi=pmi2 %s > %s\n" % (
                sub_configs["mpi_command"],
                str(self.slurm_options["ntasks"]),
                vasp_exec,
                vasp_configs["fvaspout"]
            )
        elif sub_configs['mpi_command'] == 'mpirun':
            return '\n%s -np=%s %s > %s\n' % (
                sub_configs['mpi_command'],
                str(self.slurm_options['ntasks']),
                vasp_exec,
                vasp_configs['fvaspout'])

    @property
    def lobster_command(self):
        """
        Returns command used to execute lobster
        """
        lobster = "/home/cbartel/shared/bin/lobster/lobster-4.1.0/lobster-4.1.0"
        return "\n%s\n" % lobster

    @property
    def bader_command(self):
        """
        Returns command used to execute bader
        """
        chgsum = "/home/cbartel/shared/bin/bader/chgsum.pl AECCAR0 AECCAR2"
        bader = "/home/cbartel/shared/bin/bader/bader CHGCAR -ref CHGCAR_sum"
        return "\n%s\n%s\n" % (chgsum, bader)

    def is_job_in_queue(self, job_name):
        """
        Returns:
            True if this job-name is already in the queue, else False

        Will prevent you from messing with directories that have running/pending jobs
        """
        scripts_dir = os.getcwd()
        fqueue = os.path.join(scripts_dir, "_".join(["q", job_name]) + ".o")
        with open(fqueue, "w") as f:
            subprocess.call(
                ["squeue", "-u", "%s" % os.getlogin(), "--name=%s" % job_name], stdout=f
            )
        names_in_q = []
        with open(fqueue) as f:
            for line in f:
                if "PARTITION" not in line:
                    names_in_q.append([v for v in line.split(" ") if len(v) > 0][2])
        os.remove(fqueue)
        if len(names_in_q) > 0:
            print(" !!! job already in queue, not messing with it")
            return True

        print(" not in queue, onward --> ")
        return False

    @property
    def prepare_directories(self):
        """
        This gets called by SubmitTools.write_sub, so you should rarely call this on its own

        A lot going on here. The objective is to prepare a set of directories for all calculations of interest to a given submission script in a given launch_dir
            - note: 1 submission script --> 1 chain of VASP calculations
            - note: 1 launch_dir --> can have > 1 submission script

        1) For each xc-calc pair, create a directory (calc_dir)
            - */launch_dir/xc-calc
            - xc-calc could be gga-loose, metagga-relax, etc.

        2) Check if that calc_dir has a converged VASP job
            - note: also checks "parents" (ie a static is labeled unconverged if its relax is unconverged)
                - parents determined by sub_configs['packing']
            - if calc and parents are converged:
                - checks sub_configs['fresh_restart']
                    - if fresh_restart = False --> label calc_dir as status='done' and move on
                    - if fresh_restart = True --> start this calc over

        3) Put */launch_dir/POSCAR into */launch_dir/xc-calc/POSCAR if there's not a POSCAR there already

        4) Check if */calc_dir/CONTCAR exists and has data in it,
            - if it does, copy */calc_dir/CONTCAR to */calc_dir/POSCAR and label status='continue' (ie continuing job)
            - if it doesn't, label status='new' (ie new job)

        5) Initialize VASPSetUp for calc_dir
            - modifies vasp_input_set with self.configs as requested in configs dictionaries (mainly vasp_configs which receives user_configs as well)

        6) If status in ['continue', 'new'],
            - check for errors using VASPSetUp
                - may remove WAVECAR/CHGCAR
                - will likely make edits to INCAR
        """

        # make copies of relevant configs dicts
        vasp_configs = self.vasp_configs.copy()
        sub_configs = self.sub_configs.copy()

        fresh_restart = sub_configs["fresh_restart"]
        launch_dir = self.launch_dir

        # determine the terminal ends of each chain of VASP calcs to be submitted
        final_xcs = self.final_xcs

        # determine how VASP jobs get chained together
        packing = sub_configs["packing"]

        print("\n\n~~~~~ starting to work on %s ~~~~~\n\n" % launch_dir)

        fpos_src = os.path.join(launch_dir, "POSCAR")

        # loop through all calculations within each chain and collect statuses
        # statuses = {final_xc : {xc_calc : status}}
        statuses = {}
        # looping through each chain
        for final_xc in final_xcs:
            job_name = ".".join(launch_dir.split("/")[-4:] + [final_xc])
            print("\nchecking if %s is in q" % job_name)
            if self.is_job_in_queue(job_name):
                continue
            statuses[final_xc] = {}

            # looping through each VASP calc in that chain
            counter = 0
            for xc_calc in packing[final_xc]:
                counter += 1

                # initialize configs that are particular to this particular VASP calc in this chain
                calc_configs = {}

                # (0) update vasp configs with the current xc and calc
                xc_to_run, calc_to_run = xc_calc.split("-")
                calc_configs["xc_to_run"] = xc_to_run
                calc_configs["calc_to_run"] = calc_to_run

                # (1) make calc_dir (or remove and remake if fresh_restart)
                calc_dir = os.path.join(launch_dir, xc_calc)
                if os.path.exists(calc_dir) and fresh_restart:
                    rmtree(calc_dir)
                if not os.path.exists(calc_dir):
                    os.mkdir(calc_dir)

                # (2) check convergence of current calc
                E_per_at = AnalyzeVASP(calc_dir).E_per_at
                convergence = True if E_per_at else False

                # (3) if converged, make sure parents have converged
                large_E_diff_between_relax_and_static = False
                if convergence:
                    # static calcs are currently the only ones that have parents that must be converged
                    if calc_to_run == "static":
                        static_energy = E_per_at
                        # that parent is a relax
                        parent_calc = "relax"
                        parent_xc_calc = "%s-%s" % (xc_to_run, parent_calc)
                        # make sure we're calculating relax and static (and not just)
                        if parent_xc_calc in packing[final_xc]:
                            parent_calc_dir = os.path.join(launch_dir, parent_xc_calc)
                            parent_energy = AnalyzeVASP(parent_calc_dir).E_per_at
                            parent_convergence = True if parent_energy else False
                            if not parent_energy:
                                print(
                                    "     %s (parent) not converged, need to continue this calc"
                                    % parent_xc_calc
                                )
                            else:
                                # if there is a large difference b/t the relax and static energy, something fishy happened, so let's start the static calc over
                                if abs(parent_energy - static_energy) > 0.2:
                                    print(
                                        "     %s (parent) and %s (child) energies differ by more than 0.2 eV/atom"
                                        % (parent_xc_calc, xc_calc)
                                    )
                                    large_E_diff_between_relax_and_static = True
                        else:
                            parent_convergence = True
                    else:
                        parent_convergence = True

                # if parents + current calc are converged, give it status = done
                if (
                    convergence
                    and parent_convergence
                    and not fresh_restart
                    and not large_E_diff_between_relax_and_static
                ):
                    print("     %s is already converged; skipping" % xc_calc)
                    status = "done"
                    statuses[final_xc][xc_calc] = status
                    continue

                # for jobs that are not DONE:

                # (4) check for POSCAR
                # flag to check whether POSCAR is newly copied (don't want to perturb already-perturbed structures)
                fpos_dst = os.path.join(calc_dir, "POSCAR")
                if os.path.exists(fpos_dst):
                    # if there is a POSCAR, make sure its not empty
                    contents = open(fpos_dst, "r").readlines()
                    # if its empty, copy the initial structure to calc_dir
                    if len(contents) == 0:
                        copyfile(fpos_src, fpos_dst)
                # if theres no POSCAR, copy the initial structure to calc_dir
                if not os.path.exists(fpos_dst):
                    copyfile(fpos_src, fpos_dst)

                # (5) check for CONTCAR. if one exists, if its not empty, and if not fresh_restart, mark this job as one to "continue" (ie later, we'll copy CONTCAR to POSCAR); otherwise, mark as NEWRUN
                fcont_dst = os.path.join(calc_dir, "CONTCAR")
                if os.path.exists(fcont_dst):
                    contents = open(fcont_dst, "r").readlines()
                    if (
                        (len(contents) > 0)
                        and not fresh_restart
                        and not large_E_diff_between_relax_and_static
                    ):
                        status = "continue"
                    else:
                        status = "new"
                else:
                    status = "new"

                statuses[final_xc][xc_calc] = status

                # set our user_configs based on our vasp_configs + our calc_configs
                # note: vasp_configs should hold the baseline vasp_configs + our user_configs
                # note: calc_configs should just hold xc_to_run and calc_to_run as of now
                user_vasp_configs_before_error_handling = {
                    **vasp_configs,
                    **calc_configs,
                }

                # (6) initialize VASPSetUp with current VASP configs for this calculation
                vsu = VASPSetUp(
                    calc_dir=calc_dir,
                    user_configs=user_vasp_configs_before_error_handling,
                )

                # (7) check for errors in continuing jobs
                incar_changes = {}
                if status in ["continue", "new"]:
                    calc_is_clean = vsu.is_clean
                    if not calc_is_clean:
                        # change INCAR based on errors and include in calc_configs
                        incar_changes = vsu.incar_changes_from_errors

                # if there are INCAR updates, add them to calc_configs
                if incar_changes:
                    incar_key = "%s_incar" % calc_to_run
                    if incar_key not in calc_configs:
                        calc_configs[incar_key] = {}
                    for setting in incar_changes:
                        calc_configs[incar_key][setting] = incar_changes[setting]

                # update our vasp_configs with any modifications to the INCAR that we made to fix errors
                user_vasp_configs = {**vasp_configs, **calc_configs}
                print("--------- may be some warnings (POTCAR ones OK) ----------")

                # (8) prepare calc_dir to launch
                vsu = VASPSetUp(calc_dir=calc_dir, user_configs=user_vasp_configs)

                vsu.prepare_calc

                print("-------------- warnings should be done ---------------")
                print("\n~~~~~ prepared %s ~~~~~\n" % calc_dir)
        return statuses

    @property
    def write_sub(self):
        """
        A lot going on here. The objective is to write a submission script for each calculation
            - each submission script will launch a chain of jobs
            - this gets a bit tricky because a submission script is executed in bash
                - it's essentially like moving to a compute node and typing each line of the submission script into the compute node's command line
                - this means we can't really use python while the submission script is being executed

        1) check if job's in queue. if it is, just return

        2) write our slurm options at the top of sub file

        3) loop through all the calculations we want to do from this launch dir
            - label them as "done", "continue", or "new"

        4) for "continue"
            - copy CONTCAR to POSCAR to save progress

        5) for "new" and "continue"
            - figure out what parent calculations to get data from
                - e.g., gga-static for metagga-relax
            - make sure that parent calculation finished without errors before passing data to next calc
                - and before running next calc
                - if a parent calc didnt finish, but we've moved onto the next job, kill the job, so we can (automatically) debug the parent calc

        6) write VASP commands

        7) if lobster_static and calc is static, write LOBSTER and BADER commands
        """

        final_xcs = self.final_xcs

        # make copies of our starting configs
        vasp_configs = self.vasp_configs.copy()
        sub_configs = self.sub_configs.copy()

        # determine which files will be passed from parent calcs to children
        files_to_inherit = sub_configs["files_to_inherit"]

        launch_dir = self.launch_dir

        vasp_command = self.vasp_command
        slurm_options = self.slurm_options.copy()
        queue_manager = self.queue_manager

        # determine how jobs will be chained together
        packing = sub_configs["packing"]

        # get our statuses from when we prepared VASP input files for each directory
        # statuses = {final_xc : {xc_calc : status}}
        # e.g., statuses['metagga']['gga-relax'] = 'done'
        statuses = self.prepare_directories
        for final_xc in statuses:
            # initialize a submission script for this chain
            fsub = os.path.join(launch_dir, "sub_%s.sh" % final_xc)
            # initialize a status log file for this chain
            fstatus = os.path.join(launch_dir, "status_%s.o" % final_xc)
            # create a job name based on the launch_dir
            # e.g., formula.ID.standard.mag.final_xc
            job_name = ".".join(launch_dir.split("/")[-4:] + [final_xc])
            print("\nchecking if %s is in q" % job_name)
            if self.is_job_in_queue(job_name):
                continue
            slurm_options["job-name"] = job_name
            # if job isnt in queue already, start writing a new submission script
            with open(fsub, "w") as f:
                f.write("#!/bin/bash -l\n")
                # write the SLURM stuff at the top
                for key in slurm_options:
                    slurm_option = slurm_options[key]
                    if slurm_option:
                        f.write(
                            "%s --%s=%s\n" % (queue_manager, key, str(slurm_option))
                        )
                f.write("\n\n")
                f.write("ulimit -s unlimited\n")
                if sub_configs['mpi_command'] == 'mpirun':
                    f.write('module load impi/2018/release_multithread\n')
                # now write what is needed for the chain of VASP calcs + postprocessing
                print("\n:::: writing sub now - %s ::::" % fsub)

                # use this counter to figure if there are parents for a given calc and who those parents are
                xc_calc_counter = -1
                for xc_calc in packing[final_xc]:
                    # loop through the calculations that should be chained together for a given final_xc
                    xc_calc_counter += 1
                    # grab the status
                    status = statuses[final_xc][xc_calc]
                    xc_to_run, calc_to_run = xc_calc.split("-")
                    calc_dir = os.path.join(launch_dir, xc_calc)
                    f.write(
                        "\necho working on %s _%s_ >> %s\n" % (xc_calc, status, fstatus)
                    )
                    if status == "done":
                        # write postprocessing commands if needed
                        if vasp_configs["lobster_static"]:
                            if calc_to_run == "static":
                                if sub_configs[
                                    "force_postprocess"
                                ] or not os.path.exists(
                                    os.path.join(calc_dir, "lobsterout")
                                ):
                                    f.write(self.lobster_command)
                                if sub_configs[
                                    "force_postprocess"
                                ] or not os.path.exists(
                                    os.path.join(calc_dir, "ACF.dat")
                                ):
                                    f.write(self.bader_command)
                        f.write("echo %s is done >> %s\n" % (xc_calc, fstatus))
                    else:
                        if status == "continue":
                            # copy the CONTCAR to the POSCAR
                            f.write(
                                "cp %s %s\n"
                                % (
                                    os.path.join(calc_dir, "CONTCAR"),
                                    os.path.join(calc_dir, "POSCAR"),
                                )
                            )

                        # figure out if we need to pass data from a parent calculation
                        pass_info = False if xc_calc_counter == 0 else True
                        if pass_info:
                            parent_xc_calc = packing[final_xc][xc_calc_counter - 1]
                            src_dir = os.path.join(launch_dir, parent_xc_calc)
                            # before passing data, make sure parent has finished without crashing
                            f.write(
                                "isInFile=$(cat %s | grep -c %s)\n"
                                % (os.path.join(src_dir, "OUTCAR"), "Elaps")
                            )
                            f.write("if [ $isInFile -eq 0 ]; then\n")
                            f.write(
                                '   echo "%s is not done yet so this job is being killed" >> %s\n'
                                % (parent_xc_calc, fstatus)
                            )
                            f.write("   scancel $SLURM_JOB_ID\n")
                            f.write("fi\n")

                            # pass files that need to be inherited
                            for file_to_inherit in files_to_inherit:
                                # don't pass WAVECARs from loose since KPOINTS will change
                                if ("loose" in parent_xc_calc) and (
                                    file_to_inherit == "WAVECAR"
                                ):
                                    continue
                                fsrc = os.path.join(src_dir, file_to_inherit)
                                if file_to_inherit == "CONTCAR":
                                    fdst = os.path.join(calc_dir, "POSCAR")
                                else:
                                    fdst = os.path.join(calc_dir, file_to_inherit)
                                if file_to_inherit == "CONTCAR":
                                    if os.path.exists(fsrc):
                                        contents = open(fsrc).readlines()
                                        if len(contents) < 0:
                                            continue
                                f.write("cp %s %s\n" % (fsrc, fdst))

                        # navigate to calculation directory and run vasp
                        f.write("cd %s\n" % calc_dir)
                        f.write("%s\n" % vasp_command)

                        # include postprocessing stuff as requested
                        if vasp_configs["lobster_static"]:
                            if calc_to_run == "static":
                                if (
                                    not os.path.exists(
                                        os.path.join(calc_dir, "lobsterout")
                                    )
                                    or sub_configs["force_postprocess"]
                                ):
                                    f.write(self.lobster_command)
                                if (
                                    not os.path.exists(
                                        os.path.join(calc_dir, "ACF.dat")
                                    )
                                    or sub_configs["force_postprocess"]
                                ):
                                    f.write(self.bader_command)
                        f.write(
                            "\necho launched %s-%s >> %s\n"
                            % (xc_to_run, calc_to_run, fstatus)
                        )
        return True

    @property
    def launch_sub(self):
        """
        launch the submission script written in write_sub
            - if job is not in queue already
            - if there's something to launch
                (ie if all calcs are done, dont launch)
        """
        final_xcs = self.final_xcs

        print("     now launching sub")
        scripts_dir = os.getcwd()
        launch_dir = self.launch_dir

        # determine what keywords to look for to see if job needs to be launched
        flags_that_need_to_be_executed = self.sub_configs["execute_flags"]

        for final_xc in final_xcs:

            # identify the submission script for this chain
            fsub = os.path.join(launch_dir, "sub_%s.sh" % final_xc)
            with open(fsub) as f:
                for line in f:
                    if "job-name" in line:
                        job_name = line[:-1].split("=")[-1]
            # see if jobs in queue
            if self.is_job_in_queue(job_name):
                continue
            needs_to_launch = False
            # see if there's anything to launch
            with open(fsub) as f:
                contents = f.read()
                for flag in flags_that_need_to_be_executed:
                    if flag in contents:
                        needs_to_launch = True
            if not needs_to_launch:
                print(" !!! nothing to launch here, not launching\n\n")
                return

            # if we made it this far, launch it
            os.chdir(launch_dir)
            subprocess.call(["sbatch", "sub_%s.sh" % final_xc])
            os.chdir(scripts_dir)


def main():

    return


if __name__ == "__main__":
    sub = main()



from pymatgen.entries.computed_entries import ComputedEntry, ComputedStructureEntry
from pymatgen.entries.compatibility import MaterialsProject2020Compatibility
from pymatgen.core.structure import Structure
from pydmclab.core.comp import CompTools
from pydmclab.core.struc import StrucTools
from pydmclab.hpc.analyze import AnalyzeVASP
import os


class MPCompatibilityTools(object):
    def __init__(self, calc_dir):
        """
        Args:
            data (dict): dictionary of data to create entry
                - expects {'formula' : formula (str),
                           'E_per_at' : energy per atom (float),
                           **optional keys}
            calc_dir (str): path to directory containing calculation to process
                - note: uses calc_dir if one is provided (ie overrides data)
            temperature (int): in Kelvin, must be in [0] + list(range(300, 2100, 100))

        """

        self.calc_dir = calc_dir
        if not VASPAnalysis(calc_dir).is_converged:
            raise ValueError("Calculation in %s is not converged" % calc_dir)

    @property
    def run_type(self):
        calc_dir = self.calc_dir
        calc = os.path.split(calc_dir)[-1]
        xc = calc.split("-")[0]
        if xc == "ggau":
            return "GGA+U"
        elif xc == "gga":
            return "GGA"
        elif xc == "metagga":
            return "METAGGA"

    @property
    def initial_entry(self):
        calc_dir = self.calc_dir
        structure = Structure.from_file(os.path.join(calc_dir, "CONTCAR"))
        formula = StrucTools(structure).compact_formula
        n_atoms = CompTools(formula).n_atoms
        E_per_at = VASPAnalysis(calc_dir).E_per_at
        return ComputedStructureEntry(
            structure=structure,
            energy=E_per_at * n_atoms,
            composition=formula,
            parameters={"run_type": self.run_type},
        )

    @property
    def compatibility(self):
        return MaterialsProject2020Compatibility(
            compat_type="Advanced",
            correct_peroxide=True,
            check_potcar_hash=False,
            config_file=None,
        )

    @property
    def adjustments(self):
        return self.compatibility.get_adjustments(self.initial_entry)

    @property
    def corrected_entry(self):
        initial_entry = self.initial_entry
        return ComputedEntry(
            composition=initial_entry.composition,
            energy=initial_entry.energy,
            parameters=initial_entry.parameters,
            energy_adjustments=self.adjustments,
        )

    @property
    def corrected_E_per_at(self):
        entry = self.corrected_entry
        uncorrected_E_per_at = entry.uncorrected_energy_per_atom
        correction_per_at = entry.correction_per_atom
        return uncorrected_E_per_at + correction_per_at


def main():
    calc_dir = "/panfs/jay/groups/26/cbartel/cbartel//projects/compare_to_mp/calcs/Fe1S2/mp-1522/mp/fm/ggau-static"
    compat = MPCompatibilityTools(calc_dir=calc_dir)
    return compat


if __name__ == "__main__":
    compat = main()



from .handy import read_json, read_yaml, write_json, write_yaml, dotdict, make_project_tree, is_slurm_job_in_queue, is_calc_valid, make_sub_for_launcher
from .plotting import get_colors, set_rc_params



from pydmclab.core.comp import CompTools
from pydmclab.core.query import MPQuery
from pydmclab.core.hulls import GetHullInputData, AnalyzeHull, ParallelHulls
from pydmclab.utils.handy import read_json, write_json
from pydmclab.utils.plotting import set_rc_params, tableau_colors
from pydmclab.data.thermochem import mus_at_0K, mus_at_T

import matplotlib.pyplot as plt

import os
import numpy as np

set_rc_params()

"""
This file currently tests the following classes in ThermoTools
    - GetHullInputData
    - AnalyzeHull
    - ParallelHulls

TO-DO:
    - convert into formal "tests"
"""
# Chris B's API key for MP query
API_KEY = "N3KdATtMmcsUL94g"

# chemical system to test on
CHEMSYS = "Ca-Al-Ti-O-F"

# where to save data
DATA_DIR = os.path.join("output", "hulls", "data")
if not os.path.exists(DATA_DIR):
    os.mkdir(DATA_DIR)

# where to save figures
FIG_DIR = os.path.join("output", "hulls", "figures")
if not os.path.exists(FIG_DIR):
    os.mkdir(FIG_DIR)


def get_mp_data_for_chemsys(
    chemsys=CHEMSYS,
    only_gs=True,
    dict_key="cmpd",
    remake=False,
    data_dir=DATA_DIR,
    api_key=API_KEY,
):
    """
    Args:
        chemsys (str): chemical system to query ('-'.join([elements]))
        only_gs (bool): if True, remove non-ground state polymorphs from MP query
            - good practice to do this before doing hull analysis b/c non-gs polymorphs trivially have Ehull=Ehull_gs + dE_polymorph-gs
        dict_key (str): key to use to orient dictionary of MPQuery results
            - 'cmpd' is the default behavior in MPQuery, meaning we get a dictionary that looks like {CHEMICAL_FORMULA : {DATA}}
        remake (bool): if True, re-query MP

    Returns:
        gs (dict): dictionary of ground state data from MPQuery for chemsys
    """
    fjson = os.path.join(data_dir, "query_" + chemsys + ".json")
    if not remake and os.path.exists(fjson):
        return read_json(fjson)

    mpq = MPQuery(api_key)
    out = mpq.get_data_for_comp(comp=chemsys, only_gs=only_gs, dict_key=dict_key)
    return write_json(out, fjson)


def serial_get_hull_input_data(
    gs, formation_energy_key="Ef_mp", remake=False, data_dir=DATA_DIR, chemsys=CHEMSYS
):
    """
    Args:
        gs (dict): dictionary of ground state data from MPQuery for chemsys
            - generated with get_mp_data_for_chemsys()
        formation_energy_key (str): key to use for formation energy in gs
            - 'Ef_mp' is default behavior in MPQuery
        remake (bool): if True, re-calculate hull input data
        data_dir (str): directory to save data
        chemsys (str): chemical system to query ('-'.join([elements]))

    Returns:
        hullin (dict): dictionary of hull input data for gs
            dict of {chemical space (str) : {formula (str) : {'E' : formation energy (float),
                                                              'amts' : {el (str) : fractional amt of el in formula (float) for el in space}}
                                            for all relevant formulas including elements}
                - elements are automatically given formation energy = 0
                - chemical space is now in 'el1_el2_...' format to be jsonable
                - each "chemical space" is a convex hull that must be computed
    """
    fjson = os.path.join(data_dir, "hullin_serial_" + chemsys + ".json")
    if not remake and os.path.exists(fjson):
        return read_json(fjson)

    ghid = GetHullInputData(gs, formation_energy_key=formation_energy_key)
    return ghid.hullin_data(fjson=fjson, remake=remake)


def serial_get_hull_output_data(
    hullin, remake=False, chemsys=CHEMSYS, data_dir=DATA_DIR
):
    """
    Args:
        hullin (dict): dictionary of hull input data for gs
        remake (bool): if True, re-calculate hull output data
        chemsys (str): chemical system to query ('-'.join([elements]))
        data_dir (str): directory to save data

    Returns:
        stability data (dict) for all compounds in the specified chemical space
            {compound (str) : {'Ef' : formation energy (float),
                                'Ed' : decomposition energy (float),
                                'rxn' : decomposition reaction (str),
                                'stability' : stable (True) or unstable (False)}}
    """
    fjson = os.path.join(data_dir, "hullout_serial_" + chemsys + ".json")
    if not remake and os.path.exists(fjson):
        return read_json(fjson)

    hullout = {}
    for space in hullin:
        ah = AnalyzeHull(hullin, space)
        for cmpd in hullin[space]:
            print("\n%s" % cmpd)
            hullout[cmpd] = ah.cmpd_hull_output_data(cmpd)
    return write_json(hullout, fjson)


def parallel_get_hull_input_and_output_data(
    gs, remake=False, chemsys=CHEMSYS, data_dir=DATA_DIR, n_procs=2, fresh_restart=True
):
    """
    Args:
        gs (dict): dictionary of ground state data from MPQuery for chemsys
        remake (bool): if True, re-calculate hull input and output data
        chemsys (str): chemical system to query ('-'.join([elements]))
        data_dir (str): directory to save data
        n_procs (int): number of processors to use (could also be 'all' to use multip.cpu_count()-1 procs)
        fresh_restart (bool): if True, restart ParallelHull process from scratch

    Returns:
        stability data (dict) for all compounds in the specified chemical space
            {compound (str) : {'Ef' : formation energy (float),
                                'Ed' : decomposition energy (float),
                                'rxn' : decomposition reaction (str),
                                'stability' : stable (True) or unstable (False)}}
            written to fjson
            also writes small_spaces and hullin data resulting from ParallelHull to json
    """
    fjson = os.path.join(data_dir, "hullout_parallel_" + chemsys + ".json")
    if not remake and os.path.exists(fjson):
        return read_json(fjson)
    ph = ParallelHulls(gs, n_procs=n_procs, fresh_restart=fresh_restart)
    hullin = ph.parallel_hullin(fjson=fjson.replace("hullout", "hullin"))
    smallest_spaces = ph.smallest_spaces(
        hullin=hullin, fjson=fjson.replace("hullout", "small_spaces")
    )
    return ph.parallel_hullout(
        hullin=hullin, smallest_spaces=smallest_spaces, fjson=fjson, remake=True
    )


def plot_to_check_success(gs, serial_hullout, parallel_hullout):
    """
    Args:
        gs (dict): dictionary of ground state data from MPQuery for chemsys
        serial_hullout (dict): dictionary of hull output data for gs (run serially)
        parallel_hullout (dict): dictionary of hull output data for gs (run in parallel)

    Returns:
        compares serial and parallel hull output data (should be identica)
        compares serial and MP hull output data
            - should be identical (or very close) for unstable compounds
            - for stable compounds, MP will have Ehull = 0, whereas our code will compure Ed < 0

    """
    set_rc_params()

    fig = plt.figure(figsize=(8, 3))

    params = {
        "serial": {"m": "o", "c": "blue"},
        "parallel": {"m": "^", "c": "orange"},
    }

    cmpds = sorted(gs.keys())

    cmpds = [c for c in cmpds if CompTools(c).n_els > 1]

    mp_Ehull = [gs[c]["Ehull_mp"] for c in cmpds]

    serial_decomp = [serial_hullout[c]["Ed"] for c in cmpds]
    parallel_decomp = [parallel_hullout[c]["Ed"] for c in cmpds]

    x = mp_Ehull
    y1 = serial_decomp
    y2 = parallel_decomp

    ax1 = plt.subplot(121)

    ax1 = plt.scatter(y2, y1, edgecolor="blue", marker="o", color="white")

    # ax1 = plt.xticks(xticks[1])
    # ax1 = plt.yticks(yticks[1])
    xlim, ylim = (-0.5, 1), (-0.5, 1)
    ax1 = plt.xlabel("Ed from parallel (eV/at)")
    ax1 = plt.ylabel("Ed from serial (eV/at)")
    ax1 = plt.plot(xlim, xlim, color="black", lw=1, ls="--")
    ax1 = plt.xlim(xlim)
    ax1 = plt.ylim(ylim)

    ax2 = plt.subplot(122)
    ax2 = plt.scatter(x, y1, edgecolor="blue", marker="o", color="white")

    # ax1 = plt.xticks(xticks[1])
    # ax1 = plt.yticks(yticks[1])
    xlim, ylim = (-0.1, 1), (-1, 1)
    ax2 = plt.xlabel("Ehull from MP (eV/at)")
    ax2 = plt.ylabel("")
    ax2 = plt.plot(xlim, xlim, color="black", lw=1, ls="--")
    ax2 = plt.gca().yaxis.set_ticklabels([])
    ax2 = plt.xlim(xlim)
    ax2 = plt.ylim(ylim)

    disagreements = []
    for k in serial_hullout:
        if CompTools(k).n_els == 1:
            continue
        if serial_hullout[k]["stability"] and (gs[k]["Ehull_mp"] > 0):
            disagreements.append(k)
        if not serial_hullout[k]["stability"] and (gs[k]["Ehull_mp"] == 0):
            disagreements.append(k)

        if (gs[k]["Ehull_mp"] != 0) and (
            np.round(serial_hullout[k]["Ed"], 3) != np.round(gs[k]["Ehull_mp"], 3)
        ):
            disagreements.append(k)

    for k in disagreements:
        print("\n%s" % k)
        print("my rxn = %s" % serial_hullout[k]["rxn"])
        print("my hull = %.3f" % serial_hullout[k]["Ed"])
        print("mp hull = %.3f" % gs[k]["Ehull_mp"])

    # plt.show()

    fig.savefig(os.path.join(FIG_DIR, "pd_demo_check.png"))


def main():
    # if True, re-grab data from MP
    remake_query = True
    # if True, re-calculate hull input data
    remake_serial_hullin = True
    # if True, re-calculate hull output data
    remake_serial_hullout = True
    # if True, re-calculate hull output data in parallel
    remake_parallel_hullout = True
    # if True, generate figure to check results
    remake_hull_figure_check = True

    # if True, test chemical potential stuff

    remake_mus_figure_check = True

    # MP query for CHEMSYS
    gs = get_mp_data_for_chemsys(CHEMSYS, remake=remake_query)

    # hull input data for CHEMSYS
    hullin = serial_get_hull_input_data(gs, remake=remake_serial_hullin)

    # hull output data for CHEMSYS
    hullout = serial_get_hull_output_data(hullin, remake=remake_serial_hullout)

    # hull output data for CHEMSYS (generated using parallelization)
    p_hullout = parallel_get_hull_input_and_output_data(
        gs, remake=remake_parallel_hullout
    )

    # generate a graph that compares serial vs parallel hull output and also compares ThermoTools hull output to MP hull data
    if remake_hull_figure_check:
        # %%
        plot_to_check_success(gs, hullout, p_hullout)
        # %%

    return gs, hullin, hullout, p_hullout


if __name__ == "__main__":

    # MP Query --> hull input data (serial) --> hull output data (serial) --> hull output data (parallel)
    gs, hullin, hullout, p_hullout = main()



from pydmclab.core.mag import MagTools
from pydmclab.core.struc import StrucTools
from pydmclab.hpc.analyze import AnalyzeVASP, VASPOutputs
from pydmclab.data.configs import load_vasp_configs
from pydmclab.utils.handy import read_yaml, write_yaml, dotdict

import os
import warnings
from shutil import copyfile

from pymatgen.io.vasp.sets import MPRelaxSet, MPScanRelaxSet
from pymatgen.core.structure import Structure
from pymatgen.io.vasp.inputs import Kpoints
from pymatgen.io.lobster.inputs import Lobsterin

"""
Holey Moley, getting pymatgen to find your POTCARs is not trivial...
Here's the workflow I used:
    1) Download potpaw_LDA_PBE_52_54_orig.tar.gz from VASP
    2) Extract the tar into a directory, we'll call it FULL_PATH/bin/pp
    3) Download potpaw_PBE.tgz from VASP
    4) Extract the tar INSIDE a directory: FULL_PATH/bin/pp/potpaw_PBE
    5) $ cd FULL_PATH/bin
    6) $ pmg config -p FULL_PATH/bin/pp pymatgen_pot
    7) $ pmg config --add PMG_VASP_PSP_DIR FULL_PATH/bin/pymatgen_pot
    8) $ pmg config --add PMG_DEFAULT_FUNCTIONAL PBE_54
    
Now that this has been done, new users must just do:
    1) $ pmg config --add PMG_VASP_PSP_DIR FULL_PATH/bin/pymatgen_pot
    2) $ pmg config --add PMG_DEFAULT_FUNCTIONAL PBE_54

"""


class VASPSetUp(object):
    """
    Use to write VASP inputs for a single VASP calculation
        - a calculation here might be defined as the same:
            - initial structure
            - initial magnetic configurations
            - input settings (INCAR, KPOINTS, POTCAR)
            - etc.

    Also changes inputs based on errors that are encountered

    Note that we rarely need to call this class directly
        - instead we'll manage things through pydmc.hpc.submit.SubmitTools and pydmc.hpc.launch.LaunchTools
    """

    def __init__(
        self,
        calc_dir,
        user_configs={},
        vasp_configs_yaml=os.path.join(
            os.getcwd(),
            "_vasp_configs.yaml",
        ),
        refresh_configs=True,
    ):
        """
        Args:
            calc_dir (os.PathLike) - directory where I want to execute VASP
                - there must be a POSCAR in calc_dir
                - the other input files will be added automatically within this code
            user_configs (dict) - user-defined configs
                - options in vasp_configs_yaml
            vasp_configs_yaml (os.PathLike) - path to yaml file with possible user_configs and default values
            refresh_configs (bool) - if True, will copy pydmc baseline configs to your local directory
                - this is useful if you've made changes to the configs files in the directory you're working in and want to start over

        Returns:
            calc_dir (os.PathLike) - directory where I want to execute VASP
            structure (pymatgen.Structure) - structure to be used for VASP calculation
                - note: raises error if no POSCAR in calc_dir
            configs (dict)
                # VASPSetUp (general)
                # these should get passed through LaunchTools
                standard: # make mp if comparing to MP data
                mag:  # **change** as desired
                magmom: # you will need to pass this for AFM calcs {'magmom' : [list of magmoms (float)]}

                # these should get passed through SubmitTools
                xc_to_run:  # gga, ggau, or metagga
                calc_to_run: # usually this is what we want unless you want an independent "static" or "loose" calc

                fun: # usually this is fine
                lobster_static: True # can make False if you don't want to run LOBSTER

                # some output files
                fvaspout: vasp.o # where vasp output goes in calc_dir for each vasp run
                fvasperrors: errors.o # where vasp errors go in calc_dir for each vasp run

                # may be desirable to pass various configs for each type of calculation
                # INCARs should be {INCAR_FLAG : value}
                # KPOINTs should be Kpoints object
                # POTCARs should be {el (str) : potcar flag (str)}

                # loose calculations
                loose_incar: {} # e.g., {'ENCUT' : 555}
                loose_kpoints: {} # e.g., {'length' : 25}
                loose_potcar: {} # e.g., {'W' : 'W_pv'}

                # relax calculations
                relax_incar: {}
                relax_kpoints: {}
                relax_potcar: {}

                # static calculations
                static_incar: {}
                static_kpoints: {}
                static_potcar: {}

                # lobster calculations
                lobster_incar: {}
                lobster_kpoints: {}
                lobster_potcar: {}

                # rarely need to change these:
                mag_override: False # if True, allows user to run nm calcs for mag systems and vice versa
                potcar_functional: PBE_54 # probably don't change
                validate_magmom: False # probably don't change
        """

        # this is where we will execute VASP
        self.calc_dir = calc_dir

        # we should have a POSCAR in calc_dir already
        # e.g., LaunchTools will set this up for you
        fpos = os.path.join(calc_dir, "POSCAR")
        if not os.path.exists(fpos):
            raise FileNotFoundError("POSCAR not found in {}".format(calc_dir))
        else:
            self.structure = Structure.from_file(fpos)

        # write a local yaml with vasp configs
        # if you don't have one or
        # if you want to "refresh" them
        if not os.path.exists(vasp_configs_yaml) or refresh_configs:
            _vasp_configs = load_vasp_configs()
            write_yaml(_vasp_configs, vasp_configs_yaml)

        # read local yaml to get baseline vasp_configs
        _vasp_configs = read_yaml(vasp_configs_yaml)

        # augment baseline vasp_configs with user_configs
        # NOTE: user_configs will overwrite any keys shared with vasp_configs
        configs = {**_vasp_configs, **user_configs}

        essential_configs = ["xc_to_run", "calc_to_run", "standard", "mag"]
        for essential in essential_configs:
            if essential not in configs.keys():
                raise KeyError("{} must be specified in user_configs".format(essential))

        # copy configs to prevent further changes
        self.configs = configs.copy()

        perturbation = self.configs["perturb_struc"]
        if perturbation:
            initial_structure = self.structure.copy()
            perturbed_structure = StrucTools(initial_structure).perturb(perturbation)
            self.structure = perturbed_structure

    @property
    def get_vasp_input(self):
        """
        Returns:
            vasp_input (pymatgen.io.vasp.sets.VaspInputSet

        Uses configs to modify pymatgen's VaspInputSets as required
        """

        # copy configs to prevent unwanted updates
        configs = self.configs.copy()

        # initialize how we're going to modify each vasp input file with configs specs
        modify_incar = configs["%s_incar" % configs["calc_to_run"]].copy()
        modify_kpoints = configs["%s_kpoints" % configs["calc_to_run"]].copy()
        modify_potcar = configs["%s_potcar" % configs["calc_to_run"]].copy()

        # initialize potcar functional
        potcar_functional = configs["potcar_functional"]

        # this should be kept off in general, gives unuseful warnings (I think)
        validate_magmom = configs["validate_magmom"]

        # tell user what they are modifying in case they are trying to match MP or other people's calculations
        if configs["standard"] and modify_incar:
            warnings.warn(
                "you are attempting to generate consistent data, but modifying things in the INCAR\n"
            )
            # print('e.g., %s' % str(modify_incar))

        if configs["standard"] and modify_kpoints:
            warnings.warn(
                "you are attempting to generate consistent data, but modifying things in the KPOINTS\n"
            )
            # print('e.g., %s' % str(modify_kpoints))

        if configs["standard"] and modify_potcar:
            warnings.warn(
                "you are attempting to generate consistent data, but modifying things in the POTCAR\n"
            )
            # print('e.g., %s' % str(modify_potcar))

        # tell user they are doing a nonmagnetic calculation for a compound w/ magnetic elements
        if MagTools(self.structure).could_be_magnetic and (configs["mag"] == "nm"):
            warnings.warn(
                "structure could be magnetic, but you are performing a nonmagnetic calculation\n"
            )

        structure = self.structure

        # add MAGMOM to structure
        if configs["mag"] == "nm":
            # if non-magnetic, MagTools takes care of this
            structure = MagTools(structure).get_nonmagnetic_structure
        elif configs["mag"] == "fm":
            # if ferromagnetic, MagTools takes care of this
            structure = MagTools(structure).get_ferromagnetic_structure
        elif "afm" in configs["mag"]:
            # if antiferromagnetic, we need to aprovide a MAGMOM
            magmom = configs["magmom"]
            if not magmom:
                raise ValueError("you must specify a magmom for an AFM calculation\n")
            if (min(magmom) >= 0) and (max(magmom) <= 0):
                raise ValueError(
                    "provided magmom that is not AFM, but you are trying to run an AFM calculation\n"
                )
            structure.add_site_property("magmom", magmom)

        # MP wants to set W_pv but we don't have that one in PBE54 (no biggie)
        """
        if configs["standard"] != "mp":
            modify_potcar["W"] = "W"
        """

        # don't mess with much if trying to match Materials Project
        if configs["standard"] == "mp":
            # use our default functional
            fun = None
            # set KPOINTS to be MP-consistent
            if not isinstance(modify_kpoints, dict):
                modify_kpoints = {}
                modify_kpoints["reciprocal_density"] = 64

        # setting DMC standards --> what to do on top of MPRelaxSet or MPScanRelaxSet (pymatgen defaults)
        if configs["standard"] == "dmc":
            # use our default functional
            fun = None

            # tweak a few INCAR settings
            # converge using forces (EDIFFG < 0)
            # stricter EDIFF
            # ISMEAR = 0 (less convergence errors)
            # fix ENCUT, ENAUG to be reasonable
            # turn off symmetry (mostly, ISYM = 0)
            dmc_standard_settings = {
                "EDIFF": 1e-6,
                "EDIFFG": -0.03,
                "ISMEAR": 0,
                "ENCUT": 520,
                "ENAUG": 1040,
                "ISYM": 0,
                "SIGMA": 0.01,
            }
            for key in dmc_standard_settings:
                if key not in modify_incar:
                    modify_incar[key] = dmc_standard_settings[key]

            # use length = 25 means reciprocal space discretization of 25 K-points per Å−1
            if configs["calc_to_run"] != "loose":
                if not modify_kpoints:
                    modify_kpoints = {"length": 25}

            # turn off +U unless we are specifying GGA+U
            if configs["xc_to_run"] != "ggau":
                if "LDAU" not in modify_incar:
                    modify_incar["LDAU"] = False

            # turn off ISPIN for nonmagnetic calculations
            if "ISPIN" not in modify_incar:
                modify_incar["ISPIN"] = 1 if configs["mag"] == "nm" else 2

        # start from MPRelaxSet for GGA or GGA+U
        if configs["xc_to_run"] in ["gga", "ggau"]:
            vaspset = MPRelaxSet

            # use custom functional (eg PBEsol) if you want
            # needs to be specified in user_configs['fun']
            # otherwise use PBE for gga, gga+u
            if "GGA" not in modify_incar:
                if fun:
                    modify_incar["GGA"] = fun.upper()
                else:
                    modify_incar["GGA"] = "PE"

            # for strict comparison to Materials Project GGA(+U) calculations, we need to use the old POTCARs
            if configs["standard"] == "mp":
                potcar_functional = "PBE"

        # start from MPScanRelaxSet for meta-GGA
        elif configs["xc_to_run"] == "metagga":

            vaspset = MPScanRelaxSet

            # use custom functional (eg SCAN) if you want
            # needs to be specified in user_configs['fun']
            # otherwise use r2SCAN for metagga
            if "METAGGA" not in modify_incar:
                if fun:
                    modify_incar["METAGGA"] = fun.upper()
                else:
                    modify_incar["METAGGA"] = "R2SCAN"

        # default "loose" relax
        if configs["calc_to_run"] == "loose":
            # use only 1 kpoint
            modify_kpoints = Kpoints()
            # make the settings a little looser
            loose_settings = {
                "ENCUT": 400,
                "ENAUG": 800,
                "ISIF": 3,
                "EDIFF": 1e-5,
                "NELM": 40,
            }
            for key in loose_settings:
                if key not in modify_incar:
                    modify_incar[key] = loose_settings[key]

        # default "static" claculation
        if configs["calc_to_run"] == "static":
            # don't optimize the geometry
            # do save things like charge density
            static_settings = {
                "LCHARG": True,
                "LREAL": False,
                "NSW": 0,
                "LORBIT": 0,
                "LVHAR": True,
                "ICHARG": 0,
                "LAECHG": True,
            }
            for key in static_settings:
                if key not in static_settings:
                    modify_incar[key] = static_settings[key]

        # make sure WAVECAR is written unless told user specified not to
        if "LWAVE" not in modify_incar:
            modify_incar["LWAVE"] = True

        # use better parallelization
        if ("NCORE" not in modify_incar) and ("NPAR" not in modify_incar):
            modify_incar["NCORE"] = 4

        # add more ionic steps
        if "NSW" not in modify_incar:
            if configs["calc_to_run"] == "static":
                modify_incar["NSW"] = 0
            else:
                modify_incar["NSW"] = 199

        # make sure spin is off for nm calculations
        if configs["mag"] == "nm":
            modify_incar["ISPIN"] = 1
        else:
            # make sure magnetization is written to OUTCAR for magnetic calcs
            modify_incar["LORBIT"] = 11

        # if we are doing LOBSTER, need special parameters
        # note: some of this gets handled later for us
        if configs["lobster_static"] and (configs["calc_to_run"] == "static"):
            if configs["standard"] != "mp":
                # want more DOS points
                # want to write charge densities
                # new NBANDS so don't want to start from WAVECAR
                lobster_incar_settings = {"NEDOS": 4000, "ISTART": 0, "LAECHG": True}
                for key in lobster_incar_settings:
                    if key not in configs["lobster_incar"]:
                        modify_incar[key] = lobster_incar_settings[key]

                for key in configs["lobster_incar"]:
                    modify_incar[key] = configs["lobster_incar"][key]

                if not configs["lobster_kpoints"]:
                    # need KPOINTS file for LOBSTER
                    modify_kpoints = {"length": 25}
                else:
                    modify_kpoints = configs["lobster_kpoints"]

        if configs["lobster_static"]:
            if configs["xc_to_run"] == "metagga":
                # gga-static will get ISYM = -1, so need to pass that to metagga relax otherwise WAVECAR from GGA doesnt help metagga
                modify_incar["ISYM"] = -1

        print("modify_incar = %s" % modify_incar)
        # initialize new VASPSet with all our settings
        vasp_input = vaspset(
            structure,
            user_incar_settings=modify_incar,
            user_kpoints_settings=modify_kpoints,
            user_potcar_settings=modify_potcar,
            user_potcar_functional=potcar_functional,
            validate_magmom=validate_magmom,
        )

        return vasp_input

    @property
    def prepare_calc(self):
        """
        Write input files (INCAR, KPOINTS, POTCAR)
        """

        configs = self.configs.copy()
        calc_dir = self.calc_dir

        vasp_input = self.get_vasp_input
        if not vasp_input:
            return None

        # write input files
        vasp_input.write_input(calc_dir)

        # for LOBSTER, use Janine George's Lobsterin approach (mainly to get NBANDS)
        if (configs["lobster_static"]) and (configs["calc_to_run"] == "static"):
            INCAR_input = os.path.join(calc_dir, "INCAR_input")
            INCAR_output = os.path.join(calc_dir, "INCAR")
            copyfile(INCAR_output, INCAR_input)
            POSCAR_input = os.path.join(calc_dir, "POSCAR")
            POTCAR_input = os.path.join(calc_dir, "POTCAR")
            lobsterin = Lobsterin.standard_calculations_from_vasp_files(
                POSCAR_input=POSCAR_input,
                INCAR_input=INCAR_input,
                POTCAR_input=POTCAR_input,
                option="standard",
            )
            lobsterin.write_lobsterin(os.path.join(calc_dir, "lobsterin"))
            lobsterin.write_INCAR(
                incar_input=INCAR_input,
                incar_output=INCAR_output,
                poscar_input=POSCAR_input,
            )

        return vasp_input

    @property
    def error_msgs(self):
        """
        Dict of {group of errors (str) : [list of error messages (str) in group]}
            - the error messages are things that VASP will write to fvaspout
            - we'll crawl fvaspout and assemble what errors made VASP fail,
                then we'll make edits to VASP calculation to clean them up for re-launch
        """
        return {
            "tet": [
                "Tetrahedron method fails for NKPT<4",
                "Fatal error detecting k-mesh",
                "Fatal error: unable to match k-point",
                "Routine TETIRR needs special values",
                "Tetrahedron method fails (number of k-points < 4)",
            ],
            "inv_rot_mat": [
                "inverse of rotation matrix was not found (increase " "SYMPREC)"
            ],
            "brmix": ["BRMIX: very serious problems"],
            "subspacematrix": ["WARNING: Sub-Space-Matrix is not hermitian in " "DAV"],
            "tetirr": ["Routine TETIRR needs special values"],
            "incorrect_shift": ["Could not get correct shifts"],
            "real_optlay": ["REAL_OPTLAY: internal error", "REAL_OPT: internal ERROR"],
            "rspher": ["ERROR RSPHER"],
            "dentet": ["DENTET"],
            "too_few_bands": ["TOO FEW BANDS"],
            "triple_product": ["ERROR: the triple product of the basis vectors"],
            "rot_matrix": ["Found some non-integer element in rotation matrix"],
            "brions": ["BRIONS problems: POTIM should be increased"],
            "pricel": ["internal error in subroutine PRICEL"],
            "zpotrf": ["LAPACK: Routine ZPOTRF failed"],
            "amin": ["One of the lattice vectors is very long (>50 A), but AMIN"],
            "zbrent": [
                "ZBRENT: fatal internal in",
                "ZBRENT: fatal error in bracketing",
            ],
            "pssyevx": ["ERROR in subspace rotation PSSYEVX"],
            "eddrmm": ["WARNING in EDDRMM: call to ZHEGV failed"],
            "edddav": ["Error EDDDAV: Call to ZHEGV failed"],
            "grad_not_orth": ["EDWAV: internal error, the gradient is not orthogonal"],
            "nicht_konv": ["ERROR: SBESSELITER : nicht konvergent"],
            "zheev": ["ERROR EDDIAG: Call to routine ZHEEV failed!"],
            "elf_kpar": ["ELF: KPAR>1 not implemented"],
            "elf_ncl": ["WARNING: ELF not implemented for non collinear case"],
            "rhosyg": ["RHOSYG internal error"],
            "posmap": ["POSMAP internal error: symmetry equivalent atom not found"],
            "point_group": ["Error: point group operation missing"],
            "ibzkpt": ["internal error in subroutine IBZKPT"],
            "bad_sym": [
                "ERROR: while reading WAVECAR, plane wave coefficients changed"
            ],
        }

    @property
    def unconverged_log(self):
        """
        checks to see if both ionic and electronic convergence have been reached
            if calculation had NELM # electronic steps, electronic convergence may not be met
            if calculation had NSW # ionic steps, ionic convergence may not be met

        returns a list, unconverged, that can have 0, 1, or 2 items

            if unconverged = []:
                the calculation either:
                    1) didn't finish (vasprun.xml not found or incomplete)
                    2) both ionic and electronic convergence were met
            if 'nelm_too_low' in unconverged:
                the calculation didn't reach electronic convergence
            if 'nsw_too_low' in unconverged:
                the calculation didn't reach ionic convergence
        """
        calc_dir = self.calc_dir
        configs = self.configs.copy()
        analyzer = AnalyzeVASP(calc_dir)
        outputs = VASPOutputs(calc_dir)
        unconverged = []

        # if calc is fully converged, return empty list (calc is done)
        if analyzer.is_converged:
            return unconverged

        # if vasprun doesnt exist, return empty list (calc errored out or didnt start yet)
        vr = outputs.vasprun
        if not vr:
            return unconverged

        # make sure last electronic loop converged in calc
        electronic_convergence = vr.converged_electronic

        # if we're relaxing the geometry, make sure last ionic loop converged
        if configs["calc_to_run"] == "relax":
            ionic_convergence = vr.converged_ionic
        else:
            ionic_convergence = True

        if not electronic_convergence:
            unconverged.append("nelm_too_low")
        if not ionic_convergence:
            unconverged.append("nsw_too_low")

        return unconverged

    @property
    def error_log(self):
        """
        Parse fvaspout for error messages

        Returns list of errors (str)
        """
        error_msgs = self.error_msgs
        out_file = os.path.join(self.calc_dir, self.configs["fvaspout"])
        errors = []
        with open(out_file) as f:
            contents = f.read()
        for e in error_msgs:
            for t in error_msgs[e]:
                if t in contents:
                    errors.append(e)
        return errors

    @property
    def is_clean(self):
        """
        True if no errors found and calc is fully converged, else False
        """
        configs = self.configs.copy()
        calc_dir = self.calc_dir
        clean = False
        if AnalyzeVASP(calc_dir).is_converged:
            clean = True
        if not os.path.exists(os.path.join(calc_dir, configs["fvaspout"])):
            clean = True
        if clean == True:
            with open(os.path.join(calc_dir, configs["fvasperrors"]), "w") as f:
                f.write("")
            return clean
        errors = self.error_log + self.unconverged_log
        if len(errors) == 0:
            return True
        with open(os.path.join(calc_dir, configs["fvasperrors"]), "w") as f:
            for e in errors:
                f.write(e + "\n")
        return clean

    @property
    def incar_changes_from_errors(self):
        """
        Automatic INCAR changes based on errors
            - note: also may remove WAVECAR and/or CHGCAR as needed

        Returns {INCAR key (str) : INCAR value (str)}

        This will get passed to VASPSetUp the next time we launch (using SubmitTools)

        These error fixes are mostly taken from custodian (https://github.com/materialsproject/custodian/blob/809d8047845ee95cbf0c9ba45f65c3a94840f168/custodian/vasp/handlers.py)
            + a few of my own fixes I've added over the years
        """
        calc_dir = self.calc_dir
        errors = self.error_log
        unconverged_log = self.unconverged_log
        chgcar = os.path.join(calc_dir, "CHGCAR")
        wavecar = os.path.join(calc_dir, "WAVECAR")

        incar_changes = {}
        if "grad_not_orth" in errors:
            incar_changes["SIGMA"] = 0.05
            if os.path.exists(wavecar):
                os.remove(wavecar)
            incar_changes["ALGO"] = "Exact"
        if "edddav" in errors:
            incar_changes["ALGO"] = "All"
            if os.path.exists(chgcar):
                os.remove(chgcar)
        if "eddrmm" in errors:
            if os.path.exists(wavecar):
                os.remove(wavecar)
            incar_changes["ALGO"] = "Normal"
        if "subspacematrix" in errors:
            incar_changes["LREAL"] = False
            incar_changes["PREC"] = "Accurate"
        if "inv_rot_mat" in errors:
            incar_changes["SYMPREC"] = 1e-8
        if "zheev" in errors:
            incar_changes["ALGO"] = "Exact"
        if "zpotrf" in errors:
            incar_changes["ISYM"] = -1
        if "zbrent" in errors:
            incar_changes["IBRION"] = 1
        if "brmix" in errors:
            incar_changes["IMIX"] = 1
        if "ibzkpt" in errors:
            incar_changes["SYMPREC"] = 1e-10
            incar_changes["ISMEAR"] = 0
            incar_changes["ISYM"] = -1
        if "posmap" in errors:
            incar_changes["SYMPREC"] = 1e-5
            incar_changes["ISMEAR"] = 0
            incar_changes["ISYM"] = -1
        if "nelm_too_low" in unconverged_log:
            incar_changes["NELM"] = 399
            incar_changes["ALGO"] = "All"
        if "nsw_too_low" in unconverged_log:
            incar_changes["NSW"] = 399
        if "real_optlay" in errors:
            incar_changes["LREAL"] = False
        if "bad_sym" in errors:
            incar_changes["ISYM"] = -1
        if "amin" in errors:
            incar_changes["AMIN"] = 0.01
        if "pricel" in errors:
            incar_changes["SYMPREC"] = 1e-8
            incar_changes["ISYM"] = 0
        return incar_changes


def main():

    return


if __name__ == "__main__":
    main()



import json, os, yaml, subprocess
from pydmclab.core.mag import MagTools
from pydmclab.core.comp import CompTools


def read_json(fjson):
    """
    Args:
        fjson (str) - file name of json to read

    Returns:
        dictionary stored in fjson
    """
    with open(fjson) as f:
        return json.load(f)


def write_json(d, fjson):
    """
    Args:
        d (dict) - dictionary to write
        fjson (str) - file name of json to write

    Returns:
        written dictionary
    """
    with open(fjson, "w") as f:
        json.dump(d, f)
    return d


def read_yaml(fyaml):
    """
    Args:
        fyaml (str) - file name of yaml to read

    Returns:
        dictionary stored in fjson
    """
    with open(fyaml) as f:
        return yaml.safe_load(f)


def write_yaml(d, fyaml):
    """
    Args:
        d (dict) - dictionary to write
        fyaml (str) - file name of yaml to write

    Returns:
        written dictionary
    """
    with open(fyaml, "w") as f:
        yaml.dump(d, f)
    return d


class dotdict(dict):
    """dot.notation access to dictionary attributes"""

    __getattr__ = dict.get
    __setattr__ = dict.__setitem__
    __delattr__ = dict.__delitem__


def is_slurm_job_in_queue(job_name, user_name="cbartel", fqueue="q.o"):
    with open(fqueue, "w") as f:
        subprocess.call(["squeue", "-u", user_name, "--name=%s" % job_name], stdout=f)
    names_in_q = []
    with open(fqueue) as f:
        for line in f:
            if "PARTITION" not in line:
                names_in_q.append([v for v in line.split(" ") if len(v) > 0][2])
    if len(names_in_q) == 0:
        return False
    else:
        return True


def make_project_tree():

    project_dir = os.getcwd()
    tree = {
        "dev": ["scripts", "data", "figures"],
        "results": ["scripts", "data", "figures"],
        "prep": ["notes", "background"],
        "products": ["drafts", "slides", "other"],
    }
    for first_layer in tree:
        next_layers = tree[first_layer]
        for n in next_layers:
            that_layer = os.path.join(project_dir, first_layer, n)
            if not os.path.exists(that_layer):
                os.makedirs(that_layer)
            freadme = os.path.join(that_layer, "README.md")
            if not os.path.exists(freadme):
                with open(freadme, "w") as f:
                    f.write("This is a placeholder for the %s directory" % that_layer)


def make_sub_for_launcher():
    flauncher_sub = os.path.join(os.getcwd(), "sub_launcher.sh")
    launch_job_name = "-".join([os.getcwd().split("/")[-2], "launcher"])
    with open(flauncher_sub, "w") as f:
        f.write("#!/bin/bash -l\n")
        f.write("#SBATCH --nodes=1\n")
        f.write("#SBATCH --ntasks=8\n")
        f.write("#SBATCH --time=4:00:00\n")
        f.write("#SBATCH --error=_log_launcher.e\n")
        f.write("#SBATCH --output=_log_launcher.o\n")
        f.write("#SBATCH --account=cbartel\n")
        f.write("#SBATCH --job-name=%s\n" % launch_job_name)
        f.write("#SBATCH --partition=msismall\n")
        f.write("\npython launcher.py\n")


def is_calc_valid(structure, standard, xc, calc, mag, magmom, mag_override):
    """
    Returns:
        True if calculation should be launched;
        False if some logic is violated
    """

    if standard == "mp":
        if xc != "ggau":
            return False

    if not mag_override:
        if mag == "nm":
            if MagTools(structure).could_be_magnetic:
                return False
        elif "afm" in mag:
            if not MagTools(structure).could_be_afm:
                return False
        elif mag == "fm":
            if not MagTools(structure).could_be_magnetic:
                return False

    if "afm" in mag:
        if not magmom:
            return False

    if xc == "metagga":
        if calc == "loose":
            return False

    return True


def eVat_to_kJmol(formula, eV_per_at):
    return CompTools(formula).n_atoms * eV_per_at * 96.485


def kJmol_to_eVat(formula, kJ_per_mol):
    return kJ_per_mol / (96.485 * CompTools(formula).n_atoms)



from pydmclab.core.comp import CompTools

from pymatgen.core.structure import Structure
from pymatgen.transformations.standard_transformations import (
    OrderDisorderedStructureTransformation,
    AutoOxiStateDecorationTransformation,
    OxidationStateDecorationTransformation,
)
from pymatgen.analysis.structure_matcher import StructureMatcher
from pymatgen.core.composition import Element, Composition
from pymatgen.core.ion import Ion
from pymatgen.symmetry.analyzer import SpacegroupAnalyzer

import os
import numpy as np


class StrucTools(object):
    """
    Purpose: to manipulate crystal structures for DFT calculations
    """

    def __init__(self, structure, ox_states=None):
        """
        Args:
            structure (Structure): pymatgen Structure object
                - if dict, assumes it is Structure.as_dict(); converts to Structure object
                - if str, assumes it is a path to a structure file, converts to Structure object
            ox_states (dict): dictionary of oxidation states {el (str) : oxidation state (int)}
                - or None

        """
        if isinstance(structure, dict):
            structure = Structure.from_dict(structure)
        if isinstance(structure, str):
            if os.path.exists(structure):
                structure = Structure.from_file(structure)
            else:
                raise ValueError(
                    "you passed a string to StrucTools > this means a path to a structure > but the path is empty ..."
                )
        self.structure = structure
        self.ox_states = ox_states

    @property
    def structure_as_dict(self):
        """

        Returns:
            dict: pymatgen Structure.as_dict()
        """
        return self.structure.as_dict()

    @property
    def compact_formula(self):
        """
        "clean" (reduced, systematic) formula (str) for structure
        """
        return CompTools(self.structure.formula).clean

    @property
    def formula(self):
        """
        pretty (unreduced formula) for structure
        """
        return self.structure.formula

    @property
    def els(self):
        """
        list of unique elements (str) in structure
        """
        return CompTools(self.compact_formula).els

    def make_supercell(self, grid):
        """
        Args:
            grid (list) - [nx, ny, nz]

        Returns:
            Structure repeated nx, ny, nz

            so to make a 1x2x3 supercell of the initial structure, use:
                supercell = StrucTools(structure).make_supercell([1, 2, 3])
        """
        structure = self.structure
        print("making supercell with grid %s\n" % str(grid))
        structure.make_supercell(grid)
        return structure

    def perturb(self, perturbation=0.1):
        """
        Args:
            perturbation (float) - distance in Angstrom to randomly perturb each atom

        Returns:
            Structure w/ perturbations
        """
        structure = self.structure
        structure.perturb(perturbation)
        return structure

    def change_occ(self, site_idx, new_occ, structure=None):
        """

        return a structure with a new occupation for some site

        Args:
            site_idx (int): index of site in structure to change
            new_occ (dict): dictionary telling me the new occupation on that site
                e.g., {'Li' : 0.5, 'Fe' : 0.5}

            structure (None or pymatgen Structure object):
                if None, start from self.structure
                else, start from structure

        Returns:
            pymatgen Structure object with new occupation
        """

        if not structure:
            structure = self.structure

        s = structure.copy()

        if np.sum(list(new_occ.values())) == 0:
            s.remove_sites([site_idx])
        else:
            s[site_idx].species = new_occ
        return s

    @property
    def decorate_with_ox_states(self):
        """
        Returns oxidation state decorated structure
            - uses Auto algorithm if no ox_states are provided
            - otherwise, applies ox_states
        """
        print("decorating with oxidation states\n")
        structure = self.structure
        ox_states = self.ox_states
        if not ox_states:
            print("     automatically\n")
            transformer = AutoOxiStateDecorationTransformation()
        else:
            transformer = OxidationStateDecorationTransformation(
                oxidation_states=ox_states
            )
            print("     using %s" % str(ox_states))
        return transformer.apply_transformation(structure)

    def get_ordered_structures(self, algo=0, decorate=True, n_strucs=1):
        """
        Args:
            algo (int) - 0 = fast, 1 = complete, 2 = best first
                - see pymatgen.transformations.standard_transformations.OrderDisorderedStructureTransformation
                - 0 usually OK
            decorate (bool) - whether to decorate with oxidation states
                - if False, self.structure must already have them
            n_strucs (int) - number of ordered structures to return

        Returns:
            dict of ordered structures {index : structure (Structure.as_dict())}
                - index = 0 has lowest Ewald energy
        """
        transformer = OrderDisorderedStructureTransformation(algo=algo)
        if decorate:
            structure = self.decorate_with_ox_states
        else:
            structure = self.structure
        return_ranked_list = n_strucs if n_strucs > 1 else False

        print("ordering disordered structures\n")
        out = transformer.apply_transformation(
            structure, return_ranked_list=return_ranked_list
        )
        out = [i["structure"] for i in out]
        # print(out[0])
        if isinstance(out, list):
            print("getting unique structures\n")
            matcher = StructureMatcher()
            groups = matcher.group_structures(out)
            out = [groups[i][0] for i in range(len(groups))]
            return {i: out[i].as_dict() for i in range(len(out))}
        else:
            return {0: out.as_dict()}

    def replace_species(self, species_mapping, n_strucs=1):
        """
        Args:
            species_mapping (dict) - {Element(el) :
                                        {Element(el1) : fraction el1,
                                                        fraction el2}}
            n_strucs (int) - number of ordered structures to return if disordered

        Returns:
            dict of ordered structures {index : structure (Structure.as_dict())}
                - index = 0 has lowest Ewald energy
        """
        structure = self.structure
        print("replacing species with %s\n" % str(species_mapping))

        disappearing_els = []
        for el_to_replace in species_mapping:
            if (len(species_mapping[el_to_replace]) == 1) and (
                list(species_mapping[el_to_replace].values())[0] == 0
            ):
                structure.remove_species(species=[el_to_replace])
                disappearing_els.append(el_to_replace)

        if disappearing_els:
            for el in disappearing_els:
                del species_mapping[el]

        if species_mapping:
            structure.replace_species(species_mapping)
        if structure.is_ordered:
            return {0: structure.as_dict()}
        else:
            structools = StrucTools(structure, self.ox_states)
            return structools.get_ordered_structures(n_strucs=n_strucs)

    def BROKEN_get_structures_with_dilute_vacancy(
        self, el_to_replace, n_strucs=1, structure=None
    ):
        """
        @TODO: revisit this


        Args:
            el_to_replace (str) - element to replace with vacancy
            n_strucs (int) - number of ordered structures to return if disordered
            structure (Structure) - structure to create vacancy in
                - if None, use self.structure

        Returns:
            dict of ordered structures {index : structure (Structure.as_dict())}
                - each structure will be missing 1 el_to_replace
        """
        if not structure:
            s = self.structure
        else:
            s = structure
        species_mapping = {
            Element(el_to_replace): {Element(el_to_replace): 1 - 1 / len(s)}
        }
        if not structure:
            return self.replace_species(species_mapping, n_strucs=n_strucs)
        else:
            return StrucTools(structure).replace_species(
                species_mapping, n_strucs=n_strucs
            )

    @property
    def spacegroup_info(self):
        """
        Returns:
            dict of spacegroup info with 'tight' or 'loose' symmetry tolerance
            e.g.,
                data['tight']['number'] returns spacegroup number with tight tolerance
                data['loose']['symbol'] returns spacegroup symbol with loose tolerance

        """
        data = {
            "tight": {"symprec": 0.01, "number": None, "symbol": None},
            "loose": {"symprec": 0.1, "number": None, "symbol": None},
        }
        for symprec in [0.01, 0.1]:
            sga = SpacegroupAnalyzer(self.structure, symprec=symprec)
            number = sga.get_space_group_number()
            symbol = sga.get_space_group_symbol()

            if symprec == 0.01:
                key = "tight"
            elif symprec == 0.1:
                key = "loose"

            data[key]["number"] = number
            data[key]["symbol"] = symbol

        return data

    def sg(self, number_or_symbol="symbol", loose_or_tight="loose"):
        """

        returns spacegroup number of symbol with loose or tight tolerance

        Args:
            number_or_symbol (str, optional): _description_. Defaults to 'symbol'.
            loose_or_tight (str, optional): _description_. Defaults to 'loose'.

        Returns:
            spacegroup number or symbol with loose or tight tolerance
        """
        sg_info = self.spacegroup_info
        return sg_info[loose_or_tight][number_or_symbol]


class SiteTools(object):
    """
    make it a little easier to get site info from structures

    """

    def __init__(self, structure, index):
        """
        Args:
            structure (Structure) - pymatgen structure
            index (int) - index of site in structure

        Returns:
            pymatgen Site object
        """
        if isinstance(structure, dict):
            structure = Structure.from_dict(structure)
        self.site = structure[index]

    @property
    def site_dict(self):
        """
        Returns:
            dict of site info (from Pymatgen)
        """
        return self.site.as_dict()

    @property
    def coords(self):
        """
        Returns:
            array of fractional coordinates for site ([x, y, z])
        """
        return self.site.frac_coords

    @property
    def magmom(self):
        """
        Returns:
            magnetic moment for site (float) or None
        """
        props = self.site.properties
        if props:
            if "magmom" in props:
                return props["magmom"]
        return None

    @property
    def is_fully_occ(self):
        """
        Returns:
            True if site is fully occupied else False
        """
        return self.site.is_ordered

    @property
    def ion(self):
        """
        Returns:
            whatever is occupying site (str)
                - could be multiple ions, multiple elements, one element, one ion, etc
        """
        return self.site.species_string

    @property
    def el(self):
        """
        Returns:
            just the element occupying the site (even if it has an oxidation state)
        """
        return CompTools(Composition(self.ion).formula).els[0]

    @property
    def ox_state(self):
        """
        Returns:
            oxidation state (float) of site
        """
        if self.is_fully_occ:
            return self.site_dict["species"][0]["oxidation_state"]
        else:
            print("cant determine ox state for partially occ site")
            return None


def main():

    return


if __name__ == "__main__":
    main()



#from .core import comp, mag, query, struc, hulls
#from .hpc import analyze, launch, submit, vasp
#from .utils import handy, plotting
#from .data import configs, features, reference_energies
#from .dev import energies, entries, grand, reactions



import numpy as np
import os, json

from pydmclab.core.comp import CompTools
from pydmclab.core.query import MPQuery
from pydmclab.utils.handy import write_json, read_json

this_dir, this_filename = os.path.split(__file__)
DATA_PATH = os.path.join(this_dir, "data")


def mus_at_0K():
    """
    These were run by Bartel in January 20223
    """
    fjson = os.path.join(DATA_PATH, "mus_from_dmc_no_corrections.json")
    if os.path.exists(fjson):
        return read_json(fjson)
    mus = {}
    d = read_json(os.path.join(DATA_PATH, "230122_dmc-mus.json"))
    for xc in d:
        if xc == "gga":
            functional = "pbe"
        elif xc == "metagga":
            functional = "r2scan"
        mus[functional] = d[xc]
    return write_json(mus, fjson)


def mus_at_T():
    """
    These come from Bartel 2018 Nat Comm
    """
    with open(os.path.join(DATA_PATH, "elemental_gibbs_energies_T.json")) as f:
        return json.load(f)


def mp2020_compatibility_dmus():
    """
    from MP2020Compatibility (https://github.com/materialsproject/pymatgen/blob/master/pymatgen/entries/MP2020Compatibility.yaml)
    """
    fjson = os.path.join(DATA_PATH, "mp2020_compatibility_dmus.json")
    if os.path.exists(fjson):
        return read_json(fjson)
    data = {
        "U": {
            "V": -1.7,
            "Cr": -1.999,
            "Mn": -1.668,
            "Fe": -2.256,
            "Co": -1.638,
            "Ni": -2.541,
            "W": -4.438,
            "Mo": -3.202,
        },
        "anions": {
            "O": -0.687,
            "S": -0.503,
            "F": -0.462,
            "Cl": -0.614,
            "Br": -0.534,
            "I": -0.379,
            "N": -0.361,
            "Se": -0.472,
            "Si": 0.071,
            "Sb": -0.192,
            "Te": -0.422,
            "H": -0.179,
        },
        "peroxide": {"O": -0.465},
        "superoxide": {"O": -0.161},
    }

    return write_json(data, fjson)


def mus_from_mp_no_corrections():
    """
    Last collected Dec 2022 from old API

    Returns:
        _type_: _description_
    """
    fjson = os.path.join(DATA_PATH, "mus_from_mp_no_corrections.json")
    if os.path.exists(fjson):
        return read_json(fjson)

    mus = mus_at_0K()

    mp_pbe_mus = mus["mp"]["pbe"]

    mpq = MPQuery(api_key="N3KdATtMmcsUL94g")

    mp_mus = {}
    for el in mp_pbe_mus:
        print(el)
        my_mu = mp_pbe_mus[el]
        el += "1"
        query = mpq.get_data_for_comp(el, only_gs=True)

        mp_mu = query[el]["E_mp"]
        mp_mus[el[:-1]] = mp_mu

    return write_json(mp_mus, fjson)


def ssub():
    fjson = os.path.join(DATA_PATH, "ssub.json")
    if os.path.exists(fjson):
        return read_json(fjson)
    data = {}
    with open(os.path.join(DATA_PATH, "ssub.dat")) as f:
        for line in f:
            if "cmpd" in line:
                continue
            cmpd, H = line[:-1].split(" ")
            cmpd = CompTools(cmpd).clean
            if len(CompTools(cmpd).els) > 1:
                if cmpd not in data:
                    data[cmpd] = H
                else:
                    if H < data[cmpd]:
                        data[cmpd] = H
    return write_json(data, fjson)


def mus_from_bartel2019_npj():
    fjson = os.path.join(DATA_PATH, "mus_from_bartel2019_npj.json")
    if os.path.exists(fjson):
        return read_json(fjson)

    import pandas as pd

    df = pd.read_csv(os.path.join(DATA_PATH, "bartel2019_npj_reference-energies.csv"))

    els = df.el.values
    gga = df.PBE.values
    gga_fit = df["PBE+"].values
    metagga = df.SCAN.values
    metagga_fit = df["SCAN+"].values

    xcs = {
        "pbe": gga,
        "pbe_fit": gga_fit,
        "scan": metagga,
        "scan_fit": metagga_fit,
    }

    data = {xc: {el: xcs[xc][i] for i, el in enumerate(els)} for xc in xcs}
    return write_json(data, fjson)


def main():
    mus_from_bartel2019_npj()
    ssub()
    mus_from_mp_no_corrections()
    mus_at_0K()
    mus_at_T()


if __name__ == "__main__":
    main()



#from .reference_energies import mus_at_T, mus_at_0K, mp2020_compatibility_dmus, mus_from_mp_no_corrections
#from .features import atomic_masses
#from .configs import load_vasp_configs, load_launch_configs, load_slurm_configs, load_sub_configs, load_partition_configs, load_batch_vasp_analysis_configs
#from .thermochem import ssub


import numpy as np
import os, json

this_dir, this_filename = os.path.split(__file__)
DATA_PATH = os.path.join(this_dir, "data")

def atomic_masses():
    with open(os.path.join(DATA_PATH, "atomic_masses.json")) as f:
        return json.load(f)



from pydmclab.core.struc import StrucTools, SiteTools
import itertools
import random
from pymatgen.core.structure import Structure
from pymatgen.transformations.site_transformations import (
    ReplaceSiteSpeciesTransformation,
)
from pymatgen.analysis.structure_matcher import StructureMatcher


"""
Using enumlib:
    - install enumlib from https://github.com/msg-byu/enumlib
    - for MAC
        - install xcode ($ xcode-select --install)
        - install gfortran (https://github.com/fxcoudert/gfortran-for-macOS/releases)
    - follow enumlib instructions
    - mkdir */enumlib.bin
    - move */enumlib/src/enum.x and */enumlib/aux_src/makeStr.py to */enumlib/bin
    - add #! /usr/bin/env python to top of */enumlib/bin/makeStr.py
    - add */enumlib/bin to PATH (PATH=$PATH:*/enumlib/bin)
"""


class MagTools(object):
    def __init__(
        self,
        structure,
        max_afm_combos=100,
        afm_spins=(-5, 5),
        fm_spins=(0.6, 5),
        randomize_afm=True,
        treat_as_nm=[],
    ):
        """
        Args:
            structure (Structure): pymatgen Structure object
            max_afm_combos (int): maximum number of AFM spin configurations to generate
            afm_spins (tuple): low and high spin for AFM initialization
            fm_spins (tuple): zero and non-zero spin for FM initialization
            randomize_afm (bool): randomize AFM spin configurations
                - randomization occurs in two different steps
                    - when AFM combinations are enumerated,
                        if the number of combinations is greater than max_afm_combos,
                        then we randomly select max_afm_combos combinations
                    - when the unique AFM structures are generated,
                        we randomly order them to avoid having the same "kinds" of AFM orderings always appearing first
                    - random seeds are being used so both operations should be deterministic
            treat_as_nm (list): list of elements to treat as non-magnetic
                - e.g., if you want to only explore various initial configurations for other element(s)
        """
        if isinstance(structure, dict):
            structure = Structure.from_dict(structure)
        structure.remove_oxidation_states()
        self.structure = structure
        self.max_afm_combos = max_afm_combos
        self.fm_spins = fm_spins
        self.afm_spins = afm_spins
        self.randomize_afm = randomize_afm
        self.treat_as_nm = treat_as_nm

    @property
    def magnetic_ions(self):
        """
        Aggregated from MP + matminer (**propably a better list to use somewhere**)

        from_MP = https://github.com/materialsproject/pymatgen/blob/master/pymatgen/analysis/magnetism/default_magmoms.yaml
        from_matminer = https://github.com/hackingmaterials/matminer/blob/main/matminer/featurizers/composition/element.py
        """
        from_matminer = [
            "Ti",
            "V",
            "Cr",
            "Mn",
            "Fe",
            "Co",
            "Ni",
            "Cu",
            "Nb",
            "Mo",
            "Tc",
            "Ru",
            "Rh",
            "Pd",
            "Ag",
            "Ta",
            "W",
            "Re",
            "Os",
            "Ir",
            "Pt",
        ]
        from_MP = [
            "Co",
            "Cr",
            "Fe",
            "Mn",
            "Mo",
            "Ni",
            "V",
            "W",
            "Ce",
            "Eu",
            "Ti",
            "V",
            "Cr",
            "Mn",
            "Fe",
            "Co",
            "Ni",
            "Cu",
            "Pr",
            "Nd",
            "Pm",
            "Sm",
            "Gd",
            "Tb",
            "Dy",
            "Ho",
            "er",
            "Tm",
            "Yb",
            "Np",
            "Ru",
            "Os",
            "Ir",
            "U",
        ]
        magnetic = list(set(from_matminer + from_MP))
        treat_as_nm = self.treat_as_nm
        return sorted([el for el in magnetic if el not in treat_as_nm])

    @property
    def magnetic_ions_in_struc(self):
        """
        list of elements (str) in structure that are magnetic
        """
        els = StrucTools(self.structure).els
        magnetic_ions = self.magnetic_ions
        return sorted(list(set([el for el in els if el in magnetic_ions])))

    @property
    def could_be_magnetic(self):
        """
        True if any magnetic ions are in structure else False
        """
        return True if len(self.magnetic_ions_in_struc) > 0 else False

    @property
    def could_be_afm(self):
        """
        True if there are at least two magnetic sites in structure
        """
        if not self.could_be_magnetic:
            return False

        magnetic_ions = self.magnetic_ions_in_struc
        magnetic_sites = 0
        structure = self.structure
        for idx in range(len(structure)):
            el = SiteTools(structure, idx).el
            if el in magnetic_ions:
                magnetic_sites += 1
        if magnetic_sites > 1:
            if magnetic_sites % 2 == 0:
                return True

    @property
    def get_nonmagnetic_structure(self):
        """
        Returns nonmagnetic Structure with magmom of zeros
        """
        s = self.structure
        magmom = [0 for i in range(len(s))]
        s_tmp = s.copy()
        s_tmp.add_site_property("magmom", magmom)
        return s_tmp

    @property
    def get_ferromagnetic_structure(self):
        """
        Returns Structure with all magnetic ions ordered ferromagnetically
            - nonmagnetic ions are given spin = spins[0] (default: 0.6)
            - magnetic ions are given spin = spins[1] (default: 5)
        """
        spins = self.fm_spins
        magnetic_ions_in_struc = self.magnetic_ions_in_struc
        if len(magnetic_ions_in_struc) == 0:
            return None
        s = self.structure
        # magnetic_sites = [i for i in range(len(s)) if s[i].species_string in magnetic_ions_in_struc]
        magmom = [
            spins[0] if s[i].species_string not in magnetic_ions_in_struc else spins[1]
            for i in range(len(s))
        ]
        s_tmp = s.copy()
        s_tmp.add_site_property("magmom", magmom)
        return s_tmp

    @property
    def get_antiferromagnetic_structures(self):
        """
        This is a chaotic way to get antiferromagnetic configurations
            - but it doesn't require enumlib interaction with pymatgen
            - it seems reasonably efficient, might break down for large/complex structures
            - note 1: it has no idea which configurations are "most likely" to be low energy
            - note 2: it may require execution on MSI compute nodes

        Basic workflow:
            - start from the NM structure
            - for all sites containing ions in magnetic_ions
                - generate all possible combinations of 0 (spin down) or 1 (spin up) for each site
                    - if I had four sites w/ mag ions this might be: [(0,0,0,1), (0,0,1,1), ...]
                - retain only the combinations that have average = 0.5 (ie half spin down, half spin up)
            - now apply all these combinations to the structure
                - generate a new structure for each combination that puts max(spin) on sites with 1 and min(spin) on sites with 0
            - now figure out which newly generated structures are symmetrically distinct
                - change the identities of sites that are spin up/down using oxidation state surrogate
                    - these ox states aren't physically meaningful, just a placeholder
                    - spin up: 8+, spin down: 8-
            - now use StructureMatcher to find unique structures to return

        Returns:
            list of unique Structure objects with antiferromagnetic ordering
                - exhaustive if len(combos) <= max_combos
                - no idea which are most likely to be low energy
                - reasonable to randomly sample if a very large list

            in general, probably unnecessary to calculate more than ~2-10 of these per structure
        """

        # parameters that could be args...
        spins = self.afm_spins  # magnitudes of high/low spin
        max_combos = (
            self.max_afm_combos
        )  # max number of combinations to try (for cases with very many possible combos)

        # which ions in structure are magnetic:
        magnetic_ions_in_struc = self.magnetic_ions_in_struc
        if len(magnetic_ions_in_struc) == 0:
            return None

        strucs_with_magmoms = []
        s = self.get_nonmagnetic_structure

        # get sites w/ magnetic ions
        magnetic_sites = [
            i for i in range(len(s)) if s[i].species_string in magnetic_ions_in_struc
        ]

        print("%i magnetic sites" % len(magnetic_sites))
        # enumerate all possible ways to yield afm ordering
        combos = itertools.product(range(len(spins)), repeat=len(magnetic_sites))

        combos = [c for c in combos if sum(c) == 0.5 * len(magnetic_sites)]
        print("%i afm combos" % len(combos))

        # randomly reduce list if too big for practical usage
        if len(combos) > max_combos:
            combos = random.Random(0).sample(combos, max_combos)
            print("reduced to: %i afm combos" % len(combos))

        # decorate structures w/ magmoms for all afm orderings
        for j in range(len(combos)):
            c = combos[j]
            magnetic_moments = [spins[c[i]] for i in range(len(c))]
            site_idxs_to_magmom = dict(zip(magnetic_sites, magnetic_moments))
            for i in range(len(s)):
                if i not in magnetic_sites:
                    site_idxs_to_magmom[i] = 0
            magmom = [site_idxs_to_magmom[i] for i in range(len(s))]
            s_tmp = s.copy()
            s_tmp.add_site_property("magmom", magmom)
            # print(s_tmp.site_properties)
            strucs_with_magmoms.append(s_tmp)
        print("made strucs")

        # replace spin-up and spin-down sites with new species for symmetry matching
        fake_strucs = []
        for struc in strucs_with_magmoms:
            magmom = struc.site_properties["magmom"]
            spin_up = [i for i in magnetic_sites if magmom[i] == spins[1]]
            spin_down = [i for i in magnetic_sites if magmom[i] == spins[0]]
            indices_species_map = {
                idx: struc[idx].species_string + "8+"
                if idx in spin_up
                else struc[idx].species_string + "8-"
                for idx in spin_up + spin_down
            }
            rsst = ReplaceSiteSpeciesTransformation(indices_species_map)
            s_tmp = rsst.apply_transformation(struc)
            fake_strucs.append(s_tmp)
        fake_strucs_dict = dict(zip(list(range(len(fake_strucs))), fake_strucs))

        matcher = StructureMatcher()
        groups = matcher.group_structures(
            [fake_strucs_dict[i] for i in fake_strucs_dict]
        )
        unique_strucs = []
        for g in groups:
            s = g[0]
            s.remove_oxidation_states()
            unique_strucs.append(s)

        #        print('%i unique afm structures' % len(unique_strucs))

        if self.randomize_afm:
            random.Random(1).shuffle(unique_strucs)

        print("%i unique afm structures" % len(unique_strucs))
        return unique_strucs

    @property
    def get_afm_magmoms(self):
        """
        Returns:
            dict of magmoms for each AFM ordering for a given structure
                {idx of configuration (int) : magmoms (list)}
        """
        afm_strucs = self.get_antiferromagnetic_structures
        magmoms = {}
        if afm_strucs:
            for i in range(len(afm_strucs)):
                magmoms[i] = afm_strucs[i].site_properties["magmom"]

        return magmoms


def main():

    return


if __name__ == "__main__":
    main()



from pydmclab.core.query import MPQuery
from pydmclab.core.struc import StrucTools
from pydmclab.utils.handy import read_json, write_json


import os

API_KEY = "N3KdATtMmcsUL94g"

DATA_DIR = os.path.join("output", "query")

CHEMSYS = "Li-Mn-F-S"


def demo_get_groundstate_hull_data_for_chemsys(
    chemsys=CHEMSYS,
    only_gs=True,
    dict_key="cmpd",
    remake=False,
    data_dir=DATA_DIR,
    api_key=API_KEY,
):
    """
    Args:
        chemsys (str): chemical system to query ('-'.join([elements]))
        only_gs (bool): if True, remove non-ground state polymorphs from MP query
            - good practice to do this before doing hull analysis b/c non-gs polymorphs trivially have Ehull=Ehull_gs + dE_polymorph-gs
        dict_key (str): key to use to orient dictionary of MPQuery results
            - 'cmpd' is the default behavior in MPQuery, meaning we get a dictionary that looks like {CHEMICAL_FORMULA : {DATA}}
        remake (bool): if True, re-query MP

    Returns:
        gs (dict): dictionary of ground state data from MPQuery for chemsys
    """
    fjson = os.path.join(data_dir, "query_gs_all_" + chemsys + ".json")
    if not remake and os.path.exists(fjson):
        return read_json(fjson)

    mpq = MPQuery(api_key)
    out = mpq.get_data_for_comp(comp=chemsys, only_gs=only_gs, dict_key=dict_key)
    return write_json(out, fjson)


def demo_get_all_data_for_chemsys(
    chemsys=CHEMSYS,
    only_gs=False,
    dict_key="mpid",
    remake=False,
    data_dir=DATA_DIR,
    api_key=API_KEY,
):
    """
    Args:
        chemsys (str): chemical system to query ('-'.join([elements]))
        only_gs (bool): if True, remove non-ground state polymorphs from MP query
            - good practice to do this before doing hull analysis b/c non-gs polymorphs trivially have Ehull=Ehull_gs + dE_polymorph-gs
        dict_key (str): key to use to orient dictionary of MPQuery results
            - 'cmpd' is the default behavior in MPQuery, meaning we get a dictionary that looks like {CHEMICAL_FORMULA : {DATA}}
        remake (bool): if True, re-query MP

    Returns:
        gs (dict): dictionary of ground state data from MPQuery for chemsys
    """
    fjson = os.path.join(data_dir, "query_demo_all_" + chemsys + ".json")
    if not remake and os.path.exists(fjson):
        return read_json(fjson)

    mpq = MPQuery(api_key)
    out = mpq.get_data_for_comp(comp=chemsys, only_gs=only_gs, dict_key=dict_key)
    return write_json(out, fjson)


def demo_make_sure_gs_finder_works(d_all, d_gs):

    unique_compositions = sorted(list(d_gs.keys()))

    for cmpd in unique_compositions:
        print("\n working on %s" % cmpd)
        all_IDs = [k for k in d_all if d_all[k]["cmpd"] == cmpd]
        print("found %i MP entries" % len(all_IDs))
        all_energies = [d_all[k]["Ef_mp"] for k in all_IDs]
        min_energy_from_all = min(all_energies)
        E_gs = d_gs[cmpd]["Ef_mp"]

        print(
            "Ef from all data = %.2f eV/at\nEf from gs data = %.2f eV/at"
            % (min_energy_from_all, E_gs)
        )
        if min_energy_from_all < E_gs:
            print("\n\n\n\nAHHHH\n\n\n")


def demo_entry_retrieval(mpid="mp-530748"):
    mpq = MPQuery(API_KEY)
    properties = list(mpq.typical_properties) + ["band_gap"]
    entry = mpq.get_entry_by_material_id(
        material_id=mpid,
        properties=properties,
        incl_structure=True,
        conventional=False,
        compatible_only=True,
    )
    print(entry)
    return


def demo_structure_retrieval(mpid="mp-1094961"):

    mpq = MPQuery(API_KEY)
    s = mpq.get_structure_by_material_id(material_id=mpid)

    print("\nThe formula for this structure is %s" % StrucTools(s).formula)
    print("The first site in the structure is %s" % s[0])


def demo_vaspinput_retrieval(mpid="mp-1938"):

    inputs = MPQuery(API_KEY).get_vasp_inputs(material_id=mpid)
    print("\n The %s for this calculation was %s" % ("ALGO", inputs["incar"]["ALGO"]))
    return inputs


def main():
    d_all = demo_get_all_data_for_chemsys(remake=True)
    d_gs = demo_get_groundstate_hull_data_for_chemsys(remake=True)
    demo_make_sure_gs_finder_works(d_all, d_gs)
    demo_entry_retrieval()
    demo_structure_retrieval()
    demo_vaspinput_retrieval()
    return d_all, d_gs


if __name__ == "__main__":
    d, gs = main()



from pymatgen.core.composition import Composition
import numpy as np


class CompTools(object):
    
    def __init__(self, formula):
        """
        Args: 
            formula (str) - chemical formula
        """
        self.formula = formula
    
    @property
    def clean(self):
        """
        Returns:
            formula (str) that has been:
                - sorted by elements
                - parentheses removed
                - fractions --> integers
                
        In general, we want to use this for maximum consistency
        """
        formula = Composition(self.formula).reduced_formula
        formula = Composition(formula).alphabetical_formula
        if '.' in formula:
            formula = Composition(formula).get_integer_formula_and_factor()[0]
        formula = Composition(formula).alphabetical_formula
        formula = Composition(formula).to_pretty_string()

        if (len(formula) <= 3) and (formula[-1] == '2'):
            formula = formula.replace('2', '1')
        return formula
    
    @property
    def pretty(self):
        """
        Returns:
            formula (str) that is visually pleasing (chemically meaningful)
                - note: reduces the formula
        """
        return Composition(self.formula).reduced_formula
    
    @property
    def amts(self):
        """
        Returns:
            dictionary of elements (str) and their amounts (float)
                - note: starts with "clean" formula
        """
        
        return Composition(self.clean).get_el_amt_dict()
    
    def mol_frac(self, el):
        """
        Returns:
            the molar fraction (float) of an element (str)
                - note: starts with "clean" formula
        """
        return Composition(self.clean).get_atomic_fraction(el)
    
    def stoich(self, el):
        """
        Returns:
            the stoichiometry of an element (int)
                - note: starts with "clean" formula
            e.g., if CompTools(c).clean == 'Al2Mg1O4', then CompTools(c).stoich('O') = 4
        """ 
        stoich = self.mol_frac(el) * self.n_atoms
        stoich = np.round(stoich, 0)
        return int(stoich)
    
    @property
    def chemsys(self):
        """
        Returns:
            chemical system (str) of the formula
                - sorted
                - elements (str) joined by "-"
        """
        return Composition(self.clean).chemical_system
    
    @property
    def els(self):
        """
        Returns:
            list of elements (str) in the formula
                - sorted
        """
        return list(sorted(self.chemsys.split('-')))
    
    @property
    def n_els(self):
        """
        Returns:
            number of elements (int) in the formula
        """
        return len(self.els)
    
    @property
    def n_atoms(self):
        """
        Returns:
            number of atoms (int) in the formula
                - note: starts with "clean" formula
        """
        return np.sum(list(self.amts.values()))
    
    def label_for_plot(self, el_order=None, reduce=True):
        """
        Returns:
            label (str) for plotting (includes $ for subscripts)
        """
        #formula = self.clean if reduce else self.formula
        if not el_order:
            el_order = self.els
        amts = self.amts
        label = r'$'
        for el in el_order:
            if el in amts:
                n_el = amts[el]
                if n_el == 1:
                    label += el
                elif n_el > 1:
                    if int(n_el) - n_el == 0:
                        label += el + '_{%s}' % str(int(n_el))
                    else:
                        label += el + '_{%.1f}' % float(n_el)
        label += '$'
        return label
        
    
    
    
def main():
    return 
    

if __name__ == '__main__':
    o = main()


from pydmclab.core.comp import CompTools
from pydmclab.core.struc import StrucTools

from pymatgen.ext.matproj import MPRester

# from mp_api.client import MPRester as new_MPRester

import itertools
import numpy as np


class NewMPQuery(object):
    # Chris B key = E4p0iuBO8rX7XTaPe5c4O8WO5YpikMB1

    """
    Trying to transition to new API but running into many issues
    """

    def __init__(self, api_key):
        """
        Args:
            api_key (str) - Materials Project API key

        Returns:
            self.mpr (MPRester) - Materials Project REST interface
        """

        api_key = api_key if api_key else "YOUR_API_KEY"

        self.api_key = api_key
        self.mpr = new_MPRester(api_key)

    @property
    def available_fields(self):
        return self.mpr.summary.available_fields

    @property
    def typical_fields(self):
        return [
            "formula_pretty",
            "volume",
            "nsites",
            "material_id",
            "energy_per_atom",
            "uncorrected_energy_per_atom",
            "formation_energy_per_atom",
            "energy_above_hull",
            "equilibrium_reaction_energy_per_atom",
            "decomposes_to",
            "band_gap",
            "is_magnetic",
            "symmetry",
        ]

    def get_data_for_comp(self, comp):

        all_chemsyses = None
        all_formulas = None
        if isinstance(comp, str):
            if "-" in comp:
                chemsys = comp
                all_chemsyses = []
                elements = chemsys.split("-")
                for i in range(len(elements)):
                    for els in itertools.combinations(elements, i + 1):
                        all_chemsyses.append("-".join(sorted(els)))
            else:
                all_formulas = [CompTools(comp).clean]
        elif isinstance(comp, list):
            if "-" in comp[0]:
                all_chemsyses = []
                for chemsys in comp:
                    elements = chemsys.split("-")
                    for i in range(len(elements)):
                        for els in itertools.combinations(elements, i + 1):
                            all_chemsyses.append("-".join(sorted(els)))
                all_chemsyses = sorted(list(set(all_chemsyses)))
            else:
                all_formulas = [CompTools(c).pretty for c in comp]
        if all_formulas:
            from_mp = self.mpr.summary.search(
                formula_pretty=all_formulas, fields=self.typical_fields
            )
        elif all_chemsyses:
            from_mp = self.mpr.summary.search(
                chemsys=all_chemsyses, fields=self.typical_fields
            )

        return from_mp


class MPQuery(object):
    # Chris B API KEY =
    """
    class to assist with downloading data from Materials Project

    """

    def __init__(self, api_key=None):
        """
        Args:
            api_key (str) - Materials Project API key

        Returns:
            self.mpr (MPRester) - Materials Project REST interface
        """

        api_key = api_key if api_key else "YOUR_API_KEY"

        self.api_key = api_key
        self.mpr = MPRester(api_key)

    @property
    def supported_properties(self):
        """
        Returns list of supported properties to query for MP entries in Materials Project
        """
        supported_properties = (
            "energy",
            "energy_per_atom",
            "volume",
            "formation_energy_per_atom",
            "nsites",
            "unit_cell_formula",
            "pretty_formula",
            "is_hubbard",
            "elements",
            "nelements",
            "e_above_hull",
            "hubbards",
            "is_compatible",
            "spacegroup",
            "task_ids",
            "band_gap",
            "density",
            "icsd_id",
            "icsd_ids",
            "cif",
            "total_magnetization",
            "material_id",
            "oxide_type",
            "tags",
            "elasticity",
        )

        return supported_properties

    @property
    def supported_task_properties(self):
        """
        returns list of supported properties that can be queried for any MP task
        """

        supported_task_properties = (
            "energy",
            "energy_per_atom",
            "volume",
            "formation_energy_per_atom",
            "nsites",
            "unit_cell_formula",
            "pretty_formula",
            "is_hubbard",
            "elements",
            "nelements",
            "e_above_hull",
            "hubbards",
            "is_compatible",
            "spacegroup",
            "band_gap",
            "density",
            "icsd_id",
            "cif",
        )

        return supported_task_properties

    @property
    def typical_properties(self):
        """
        A list of propreties that we often query for

        """
        typical_properties = (
            "energy_per_atom",
            "pretty_formula",
            "material_id",
            "formation_energy_per_atom",
            "e_above_hull",
            "nsites",
            "volume",
            "spacegroup.number",
        )
        return typical_properties

    @property
    def long_to_short_keys(self):
        """
        A map to nickname query properties with shorter handles
            (dict)

        So after querying 'energy_per_atom' will be a key, but this map will convert that to 'E_mp'
        """
        return {
            "energy_per_atom": "E_mp",
            "formation_energy_per_atom": "Ef_mp",
            "e_above_hull": "Ehull_mp",
            "spacegroup.number": "sg",
            "material_id": "mpid",
        }

    def get_data_for_comp(
        self,
        comp,
        properties=None,
        criteria={},
        only_gs=True,
        include_structure=True,
        supercell_structure=False,
        max_Ehull=0.1,
        max_sites_per_structure=100,
        max_strucs_per_cmpd=5,
    ):
        """
        Args:
            comp (list or str)
                can either be:
                    - a chemical system (str) of elements joined by "-"
                    - a chemical formula (str)
                can either be a list of:
                    - chemical systems (str) of elements joined by "-"
                    - chemical formulas (str)

            properties (list or None)
                list of properties to query
                    - if None, then use typical_properties
                    - if 'all', then use supported_properties

            criteria (dict or None)
                dictionary of criteria to query
                    - if None, then use {}

            only_gs (bool)
                if True, remove non-ground state polymorphs for each unique composition

            include_structure (bool)
                if True, include the structure (as a dictionary) for each entry

            supercell_structure (bool)
                only runs if include_structure = True
                if False, just retrieve the MP structure
                if not False, must be specified as [a,b,c] to make an a x b x c supercell of the MP structure

            max_Ehull (float)
                if not None, remove entries with Ehull_mp > max_Ehull

            max_sites_per_structure (int)
                if not None, remove entries with more than max_sites_per_structure sites

            max_strucs_per_cmpd (int)
                if not None, only retain the lowest energy structures for each composition until you reach max_strucs_per_cmpd

        Returns:
            {mpid : {DATA}}
        """
        key_map = self.long_to_short_keys
        if properties == "all":
            properties = self.supported_properties
        if properties == None:
            properties = self.typical_properties
        else:
            for prop in properties:
                if prop not in self.supported_properties:
                    raise ValueError("Property %s is not supported!" % prop)

        if criteria == None:
            criteria = {}

        if isinstance(comp, str):
            if "-" in comp:
                chemsys = comp
                all_chemsyses = []
                elements = chemsys.split("-")
                for i in range(len(elements)):
                    for els in itertools.combinations(elements, i + 1):
                        all_chemsyses.append("-".join(sorted(els)))

                criteria["chemsys"] = {"$in": all_chemsyses}
            else:
                formula = comp
                criteria["pretty_formula"] = {"$in": [CompTools(formula).pretty]}

        elif isinstance(comp, list):
            if "-" in comp[0]:
                all_chemsyses = []
                for chemsys in comp:
                    elements = chemsys.split("-")
                    for i in range(len(elements)):
                        for els in itertools.combinations(elements, i + 1):
                            all_chemsyses.append("-".join(sorted(els)))
                all_chemsyses = sorted(list(set(all_chemsyses)))
                criteria["chemsys"] = {"$in": all_chemsyses}
            else:
                all_formulas = [CompTools(c).pretty for c in comp]
                criteria["pretty_formula"] = {"$in": all_formulas}

        list_from_mp = self.mpr.query(criteria, properties)
        extra_keys = [k for k in list_from_mp[0] if k not in key_map]
        cleaned_list_from_mp = [
            {key_map[old_key]: entry[old_key] for old_key in key_map}
            for entry in list_from_mp
        ]
        query = []
        for i in range(len(list_from_mp)):
            query.append(
                {
                    **cleaned_list_from_mp[i],
                    **{k: list_from_mp[i][k] for k in extra_keys},
                    **{"cmpd": CompTools(list_from_mp[i]["pretty_formula"]).clean},
                }
            )

        if only_gs:

            gs = {}
            for entry in query:
                cmpd = CompTools(entry["pretty_formula"]).clean
                if cmpd not in gs:
                    gs[cmpd] = entry
                else:
                    energy_key = "E_mp" if len(CompTools(cmpd).els) == 1 else "Ef_mp"
                    Ef_stored = gs[cmpd][energy_key]
                    Ef_check = entry[energy_key]
                    if Ef_check < Ef_stored:
                        gs[cmpd] = entry
            query = [gs[k] for k in gs]
            query = {entry["mpid"]: entry for entry in query}

        else:
            query = {entry["mpid"]: entry for entry in query}

        if include_structure:
            for entry in query:
                mpid = query[entry]["mpid"]
                structure = self.get_structure_by_material_id(mpid)
                if supercell_structure:
                    if len(supercell_structure) == 3:
                        structure = StrucTools(structure).make_supercell(
                            supercell_structure
                        )
                query[entry]["structure"] = structure.as_dict()

        if max_sites_per_structure:
            query = {
                e: query[e]
                for e in query
                if query[e]["nsites"] <= max_sites_per_structure
            }

        if max_Ehull:
            if only_gs:
                query = {
                    e: query[e] for e in query if query[e]["Ehull_mp"] <= max_Ehull
                }
            else:
                cmpds = sorted(list(set([query[e]["cmpd"] for e in query])))
                mpids = [
                    e
                    for e in query
                    if query[e]["Ehull_mp"] <= max_Ehull
                    if query[e]["cmpd"] in cmpds
                ]
                query = {e: query[e] for e in mpids}

        if max_strucs_per_cmpd:
            if not only_gs:
                trimmed_query = {}
                cmpds = sorted(list(set([query[e]["cmpd"] for e in query])))
                for cmpd in cmpds:
                    mpids = [e for e in query if query[e]["cmpd"] == cmpd]
                    Ehulls = [query[e]["Ehull_mp"] for e in mpids]
                    sorted_indices = np.argsort(Ehulls)
                    relevant_ids = [mpids[i] for i in sorted_indices]
                    if len(relevant_ids) > max_strucs_per_cmpd:
                        relevant_ids = relevant_ids[:max_strucs_per_cmpd]
                    for mpid in relevant_ids:
                        trimmed_query[mpid] = query[mpid]

                return trimmed_query

        return query

    def get_entry_by_material_id(
        self,
        material_id,
        properties=None,
        incl_structure=True,
        conventional=False,
        compatible_only=True,
    ):
        """
        Args:
            material_id (str) - MP ID of entry
            properties (list) - list of properties to query
            incl_structure (bool) - whether to include structure in entry
            conventional (bool) - whether to use conventional unit cell
            compatible_only (bool) - whether to only include compatible entries (related to MP formation energies)

        Returns:
            ComputedEntry object
        """
        return self.mpr.get_entry_by_material_id(
            material_id, compatible_only, incl_structure, properties, conventional
        )

    def get_structure_by_material_id(self, material_id):
        """
        Args:
            material_id (str) - MP ID of entry

        Returns:
            Structure object
        """
        return self.mpr.get_structure_by_material_id(material_id)

    def get_incar(self, material_id):
        """
        Args:
            material_id (str) - MP ID of entry

        Returns:
            dict of incar settings
        """
        return self.mpr.query(material_id, ["input.incar"])[0]

    def get_kpoints(self, material_id):
        """
        Args:
            material_id (str) - MP ID of entry

        Returns:
            dict of kpoint settings
        """
        return self.mpr.query(material_id, ["input.kpoints"])[0][
            "input.kpoints"
        ].as_dict()

    def get_vasp_inputs(self, material_id):
        """
        Args:
            material_id (str) - MP ID of entry

        Returns:
            dict of vasp inputs
                - 'incar' : {setting (str) : value (mixed type)}
                - 'kpoints' : {'scheme' : (str), 'grid' : list of lists for 'A B C'}
                - 'potcar' : [list of TITELs]
                - 'structure' : Structure object as dict
        """

        d = self.mpr.query(material_id, ["input"])[0]["input"]
        d["kpoints"] = d["kpoints"].as_dict()
        d["kpoints"] = {
            "scheme": d["kpoints"]["generation_style"],
            "grid": d["kpoints"]["kpoints"],
        }
        d["potcar"] = [
            d["potcar_spec"][i]["titel"] for i in range(len(d["potcar_spec"]))
        ]
        d["poscar"] = self.get_structure_by_material_id(material_id).as_dict()
        del d["potcar_spec"]

        return d


def main():

    return


if __name__ == "__main__":
    mpq = main()
